# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, division, absolute_import

from struct import unpack, unpack_from, calcsize

from ..message import BackendMessage
from ....datatypes import getTypeName


class ParameterDescription(BackendMessage):
    message_id = b't'

    def __init__(self, data):
        BackendMessage.__init__(self)
        self.parameters = []
        self.parameter_count = unpack('!H', data[0:2])[0]
        if self.parameter_count == 0:
            return

        # read type pool
        # used for special types e.g. GEOMETRY, GEOGRAPHY
        user_types = []
        type_pool_count = unpack('!I', data[2:6])[0]
        pos = 6
        for _ in range(type_pool_count):
            base_type_oid = unpack('!I', data[pos:(pos + 4)])[0]
            pos += 4
            type_name = unpack_from("!{0}sx".format(data.find(b'\x00', pos) - pos), data, pos)[0]
            pos += len(type_name) + 1
            user_types.append((base_type_oid, type_name))

        # read info of each parameter
        offset = calcsize("!BIiH")
        for _ in range(self.parameter_count):
            field_info = unpack_from("!BIiH", data, pos)
            pos += offset

            if field_info[0] == 1:
                data_type_oid, data_type_name = user_types[field_info[1]]
            else:
                data_type_oid = field_info[1]
                data_type_name = getTypeName(data_type_oid, field_info[2])

            self.parameters.append({
                'data_type_oid': data_type_oid,
                'data_type_name': data_type_name,
                'type_modifier': field_info[2],
                'null_ok': field_info[3] != 1,
            })

    def __str__(self):
        return "ParameterDescription: {}".format(self.parameters)


BackendMessage.register(ParameterDescription)



# Copyright (c) 2019-2020 Micro Focus or one of its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from multiprocessing import Process
import pytest
import time

from .base import VerticaPythonIntegrationTestCase
from ... import errors


class CancelTestCase(VerticaPythonIntegrationTestCase):

    def test_cursor_cancel(self):
        # Cursor.cancel() should be not supported any more
        with self._connect() as conn:
            cursor = conn.cursor()
            with self.assertRaises(errors.NotSupportedError):
                cursor.cancel()

    def test_connection_cancel_no_query(self):
        with self._connect() as conn:
            cur = conn.cursor()
            # No query is being executed, cancel does nothing
            conn.cancel()

    @pytest.mark.timeout(30)
    def test_connection_cancel_running_query(self):
        def cancel_query(conn, delay=5):
            time.sleep(delay)
            conn.cancel()

        with self._connect() as conn:
            cur = conn.cursor()
            p1 = Process(target=cancel_query, args=(conn,))
            p1.start()
            with self.assertRaises(errors.QueryCanceled):
                long_running_query = ('select count(*) from '
                    '(select node_name from CONFIGURATION_PARAMETERS) as a cross join '
                    '(select node_name from CONFIGURATION_PARAMETERS) as b cross join '
                    '(select node_name from CONFIGURATION_PARAMETERS) as c')
                cur.execute(long_running_query)
            p1.join()


exec(CancelTestCase.createPrepStmtClass())



# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, division, absolute_import

import getpass
import uuid
from .base import VerticaPythonIntegrationTestCase


class ConnectionTestCase(VerticaPythonIntegrationTestCase):

    def tearDown(self):
        super(ConnectionTestCase, self).tearDown()
        if 'session_label' in self._conn_info:
            del self._conn_info['session_label']
        if 'autocommit' in self._conn_info:
            del self._conn_info['autocommit']

    def test_client_os_user_name_metadata(self):
        value = getpass.getuser()

        # Metadata client_os_user_name sent from client should be captured into system tables
        query = 'SELECT client_os_user_name FROM v_monitor.current_session'
        res = self._query_and_fetchone(query)
        self.assertEqual(res[0], value)

        query = 'SELECT client_os_user_name FROM v_monitor.sessions WHERE session_id=(SELECT current_session())'
        res = self._query_and_fetchone(query)
        self.assertEqual(res[0], value)

        query = 'SELECT client_os_user_name FROM v_monitor.user_sessions WHERE session_id=(SELECT current_session())'
        res = self._query_and_fetchone(query)
        self.assertEqual(res[0], value)

        query = 'SELECT client_os_user_name FROM v_internal.dc_session_starts WHERE session_id=(SELECT current_session())'
        res = self._query_and_fetchone(query)
        self.assertEqual(res[0], value)

    def test_session_label(self):
        label = str(uuid.uuid1())
        self._conn_info['session_label'] = label

        query = 'SELECT client_label FROM v_monitor.current_session'
        res = self._query_and_fetchone(query)
        self.assertEqual(res[0], label)

        query = 'SELECT client_label FROM v_monitor.sessions WHERE session_id=(SELECT current_session())'
        res = self._query_and_fetchone(query)
        self.assertEqual(res[0], label)

        query = 'SELECT client_label FROM v_monitor.user_sessions WHERE session_id=(SELECT current_session())'
        res = self._query_and_fetchone(query)
        self.assertEqual(res[0], label)

        query = 'SELECT client_label FROM v_internal.dc_session_starts WHERE session_id=(SELECT current_session())'
        res = self._query_and_fetchone(query)
        self.assertEqual(res[0], label)

    def test_autocommit_on(self):
        # Set with connection option
        self._conn_info['autocommit'] = True
        with self._connect() as conn:
            self.assertTrue(conn.autocommit)
            # Set with attribute setter
            conn.autocommit = False
            self.assertFalse(conn.autocommit)

    def test_autocommit_off(self):
        # Set with connection option
        self._conn_info['autocommit'] = False
        with self._connect() as conn:
            self.assertFalse(conn.autocommit)
            # Set with attribute setter
            conn.autocommit = True
            self.assertTrue(conn.autocommit)

exec(ConnectionTestCase.createPrepStmtClass())



# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, division, absolute_import

import os
import sys
import logging
import unittest
import inspect
import getpass
import six
if six.PY2:
    from ConfigParser import ConfigParser
elif six.PY3:
    from configparser import ConfigParser

from ...compat import as_text, as_str, as_bytes
from ...vertica.log import VerticaLogging


default_configs = {
    'log_dir': 'vp_test_log',
    'log_level': logging.WARNING,
    'host': 'localhost',
    'port': 5433,
    'user': getpass.getuser(),
    'password': '',
}


class VerticaPythonTestCase(unittest.TestCase):
    """
    Base class for all tests
    """

    @classmethod
    def _load_test_config(cls, config_list):
        test_config = {}

        # load default configurations
        for key in config_list:
            if key != 'database':
                test_config[key] = default_configs[key]

        # override with the configuration file
        confparser = ConfigParser()
        confparser.optionxform = str
        SECTION = 'vp_test_config'  # section name in the configuration file
        # the configuration file is placed in the same directory as this file
        conf_file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                 'vp_test.conf')
        confparser.read(conf_file)
        for k in config_list:
            option = 'VP_TEST_' + k.upper()
            if confparser.has_option(SECTION, option):
                test_config[k] = confparser.get(SECTION, option)

        # override again with VP_TEST_* environment variables
        for k in config_list:
            env = 'VP_TEST_' + k.upper()
            if env in os.environ:
                test_config[k] = os.environ[env]

        # data preprocessing
        # value is string when loaded from configuration file and environment variable
        if 'port' in test_config:
            test_config['port'] = int(test_config['port'])
        if 'database' in config_list and 'user' in test_config:
            test_config.setdefault('database', test_config['user'])
        if 'log_level' in test_config:
            levels = ['NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']
            if isinstance(test_config['log_level'], str):
                if test_config['log_level'] not in levels:
                    raise ValueError("Invalid value for VP_TEST_LOG_LEVEL: '{}'".format(test_config['log_level']))
                test_config['log_level'] = eval('logging.' + test_config['log_level'])
        if 'log_dir' in test_config:
            test_config['log_dir'] = os.path.join(test_config['log_dir'],
                                     'py{0}{1}'.format(sys.version_info.major, sys.version_info.minor))

        return test_config

    @classmethod
    def _setup_logger(cls, tag, log_dir, log_level):
        # Setup test logger
        # E.g. If the class is defined in tests/integration_tests/test_dates.py
        #      and test cases run under python3.7, then
        #      the log would write to $VP_TEST_LOG_DIR/py37/integration_tests/test_dates.log

        testfile = os.path.splitext(os.path.basename(inspect.getsourcefile(cls)))[0]
        logfile = os.path.join(log_dir, tag, testfile + '.log')
        VerticaLogging.setup_logging(cls.__name__, logfile, log_level, cls.__name__)
        cls.logger = logging.getLogger(cls.__name__)
        return logfile

    def setUp(self):
        self.logger.info('\n\n'+'-'*50+'\n Begin '+self.__class__.__name__+"."+self._testMethodName+'\n'+'-'*50)

    def tearDown(self):
        self.logger.info('\n'+'-'*10+' End '+self.__class__.__name__+"."+self._testMethodName+' '+'-'*10+'\n')

    # Common assertions
    def assertTextEqual(self, first, second, msg=None):
        first_text = as_text(first)
        second_text = as_text(second)
        self.assertEqual(first=first_text, second=second_text, msg=msg)

    def assertStrEqual(self, first, second, msg=None):
        first_str = as_str(first)
        second_str = as_str(second)
        self.assertEqual(first=first_str, second=second_str, msg=msg)

    def assertBytesEqual(self, first, second, msg=None):
        first_bytes = as_bytes(first)
        second_bytes = as_bytes(second)
        self.assertEqual(first=first_bytes, second=second_bytes, msg=msg)

    def assertResultEqual(self, value, result, msg=None):
        if isinstance(value, six.string_types):
            self.assertTextEqual(first=value, second=result, msg=msg)
        else:
            self.assertEqual(first=value, second=result, msg=msg)

    def assertListOfListsEqual(self, list1, list2, msg=None):
        self.assertEqual(len(list1), len(list2), msg=msg)
        for l1, l2 in zip(list1, list2):
            self.assertListEqual(l1, l2, msg=msg)



# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, division, absolute_import

from .base import VerticaPythonIntegrationTestCase


class LoadBalanceTestCase(VerticaPythonIntegrationTestCase):
    def setUp(self):
        super(LoadBalanceTestCase, self).setUp()
        self._host = self._conn_info['host']
        self._port = self._conn_info['port']

    def tearDown(self):
        self._conn_info['host'] = self._host
        self._conn_info['port'] = self._port
        self._conn_info['connection_load_balance'] = False
        self._conn_info['backup_server_node'] = []

        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("SELECT set_load_balance_policy('NONE')")
            cur.execute("DROP TABLE IF EXISTS test_loadbalance")
        super(LoadBalanceTestCase, self).tearDown()

    def get_node_num(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("SELECT count(*) FROM nodes WHERE node_state='UP'")
            db_node_num = cur.fetchone()[0]
            return db_node_num

    def test_loadbalance_option_disabled(self):
        if 'connection_load_balance' in self._conn_info:
            del self._conn_info['connection_load_balance']
        self.assertConnectionSuccess()

        self._conn_info['connection_load_balance'] = False
        self.assertConnectionSuccess()

    def test_loadbalance_random(self):
        self.require_DB_nodes_at_least(3)
        self._conn_info['connection_load_balance'] = True
        rowsToInsert = 3 * self.db_node_num

        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("SELECT set_load_balance_policy('RANDOM')")
            cur.execute("DROP TABLE IF EXISTS test_loadbalance")
            cur.execute("CREATE TABLE test_loadbalance (n varchar)")
            # record which node the client has connected to
            for i in range(rowsToInsert):
                with self._connect() as conn1:
                    cur1 = conn1.cursor()
                    cur1.execute("INSERT INTO test_loadbalance (SELECT node_name FROM sessions "
                                 "WHERE session_id = (SELECT current_session()))")
                    conn1.commit()

            cur.execute("SELECT count(DISTINCT n)>1 FROM test_loadbalance")
            res = cur.fetchone()
            self.assertTrue(res[0])

    def test_loadbalance_none(self):
        # Client turns on connection_load_balance but server is unsupported
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("SELECT set_load_balance_policy('NONE')")
        self._conn_info['connection_load_balance'] = True

        # Client will proceed with the existing connection with initiator
        self.assertConnectionSuccess()

        # Test for multi-node DB
        self.require_DB_nodes_at_least(3)
        rowsToInsert = 3 * self.db_node_num

        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("DROP TABLE IF EXISTS test_loadbalance")
            cur.execute("CREATE TABLE test_loadbalance (n varchar)")
            # record which node the client has connected to
            for i in range(rowsToInsert):
                with self._connect() as conn1:
                    cur1 = conn1.cursor()
                    cur1.execute("INSERT INTO test_loadbalance (SELECT node_name FROM sessions "
                                 "WHERE session_id = (SELECT current_session()))")
                    conn1.commit()

            cur.execute("SELECT count(DISTINCT n)=1 FROM test_loadbalance")
            res = cur.fetchone()
            self.assertTrue(res[0])

    def test_loadbalance_roundrobin(self):
        self.require_DB_nodes_at_least(3)
        self._conn_info['connection_load_balance'] = True
        rowsToInsert = 3 * self.db_node_num

        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("SELECT set_load_balance_policy('ROUNDROBIN')")
            cur.execute("DROP TABLE IF EXISTS test_loadbalance")
            cur.execute("CREATE TABLE test_loadbalance (n varchar)")
            # record which node the client has connected to
            for i in range(rowsToInsert):
                with self._connect() as conn1:
                    cur1 = conn1.cursor()
                    cur1.execute("INSERT INTO test_loadbalance (SELECT node_name FROM sessions "
                                 "WHERE session_id = (SELECT current_session()))")
                    conn1.commit()

            cur.execute("SELECT count(n)=3 FROM test_loadbalance GROUP BY n")
            res = cur.fetchall()
            # verify that all db_node_num nodes are represented equally
            self.assertEqual(len(res), self.db_node_num)
            for i in res:
                self.assertEqual(i, [True])

    def test_failover_empty_backup(self):
        # Connect to primary server
        if 'backup_server_node' in self._conn_info:
            del self._conn_info['backup_server_node']
        self.assertConnectionSuccess()
        self._conn_info['backup_server_node'] = []
        self.assertConnectionSuccess()

        # Set primary server to invalid host and port
        self._conn_info['host'] = 'invalidhost'
        self._conn_info['port'] = 9999

        # Fail to connect to primary server
        self.assertConnectionFail()

    def test_failover_one_backup(self):
        # Set primary server to invalid host and port
        self._conn_info['host'] = 'invalidhost'
        self._conn_info['port'] = 9999

        # One valid address in backup_server_node: port is an integer
        self._conn_info['backup_server_node'] = [(self._host, self._port)]
        self.assertConnectionSuccess()

        # One valid address in backup_server_node: port is a string
        self._conn_info['backup_server_node'] = [(self._host, str(self._port))]
        self.assertConnectionSuccess()

        # One invalid address in backup_server_node: DNS failed, Name or service not known
        self._conn_info['backup_server_node'] = [('invalidhost2', 8888)]
        self.assertConnectionFail()

        # One invalid address in backup_server_node: DNS failed, Name or service not known
        self._conn_info['backup_server_node'] = [('123.456.789.123', 8888)]
        self.assertConnectionFail()

        # One invalid address in backup_server_node: DNS failed, Address family for hostname not supported
        self._conn_info['backup_server_node'] = [('fd76:6572:7469:6361:0:242:ac11:4', 8888)]
        self.assertConnectionFail()

        # One invalid address in backup_server_node: Wrong port, connection refused
        self._conn_info['backup_server_node'] = [(self._host, 8888)]
        self.assertConnectionFail()

    def test_failover_multi_backup(self):
        # Set primary server to invalid host and port
        self._conn_info['host'] = 'invalidhost'
        self._conn_info['port'] = 9999

        # One valid and two invalid addresses in backup_server_node
        self._conn_info['backup_server_node'] = [(self._host, self._port), 'invalidhost2', 'foo']
        self.assertConnectionSuccess()
        self._conn_info['backup_server_node'] = ['foo', (self._host, self._port), ('123.456.789.1', 888)]
        self.assertConnectionSuccess()
        self._conn_info['backup_server_node'] = ['foo', ('invalidhost2', 8888), (self._host, self._port)]
        self.assertConnectionSuccess()

        # Three invalid addresses in backup_server_node
        self._conn_info['backup_server_node'] = ['foo', (self._host, 9999), ('123.456.789.1', 888)]
        self.assertConnectionFail()

    def test_failover_backup_format(self):
        # Set primary server to invalid host and port
        self._conn_info['host'] = 'invalidhost'
        self._conn_info['port'] = 9999

        err_msg = 'Connection option "backup_server_node" must be a list'
        self._conn_info['backup_server_node'] = (self._host, self._port)
        self.assertConnectionFail(TypeError, err_msg)

        err_msg = ('Each item of connection option "backup_server_node"'
                   r' must be a host string or a \(host, port\) tuple')
        self._conn_info['backup_server_node'] = [9999]
        self.assertConnectionFail(TypeError, err_msg)
        self._conn_info['backup_server_node'] = [(self._host, self._port, 'foo', 9999)]
        self.assertConnectionFail(TypeError, err_msg)

        err_msg = 'Host must be a string: invalid value: .*'
        self._conn_info['backup_server_node'] = [(9999, self._port)]
        self.assertConnectionFail(TypeError, err_msg)
        self._conn_info['backup_server_node'] = [(9999, 'port_num')]
        self.assertConnectionFail(TypeError, err_msg)

        err_msg = 'Port must be an integer or a string: invalid value: .*'
        self._conn_info['backup_server_node'] = [(self._host, 5433.0022)]
        self.assertConnectionFail(TypeError, err_msg)

        err_msg = r'Port .* is not a valid string: invalid literal for int\(\) with base 10: .*'
        self._conn_info['backup_server_node'] = [(self._host, 'port_num')]
        self.assertConnectionFail(ValueError, err_msg)
        self._conn_info['backup_server_node'] = [(self._host, '5433.0022')]
        self.assertConnectionFail(ValueError, err_msg)

        err_msg = 'Invalid port number: .*'
        self._conn_info['backup_server_node'] = [(self._host, -1000)]
        self.assertConnectionFail(ValueError, err_msg)
        self._conn_info['backup_server_node'] = [(self._host, 66000)]
        self.assertConnectionFail(ValueError, err_msg)
        self._conn_info['backup_server_node'] = [(self._host, '-1000')]
        self.assertConnectionFail(ValueError, err_msg)
        self._conn_info['backup_server_node'] = [(self._host, '66000')]
        self.assertConnectionFail(ValueError, err_msg)

    def test_failover_with_loadbalance_roundrobin(self):
        self.require_DB_nodes_at_least(3)

        # Set primary server to invalid host and port
        self._conn_info['host'] = 'invalidhost'
        self._conn_info['port'] = 9999
        self.assertConnectionFail()

        self._conn_info['backup_server_node'] = [('invalidhost2', 8888), (self._host, self._port)]
        self.assertConnectionSuccess()

        self._conn_info['connection_load_balance'] = True
        rowsToInsert = 3 * self.db_node_num

        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("SELECT set_load_balance_policy('ROUNDROBIN')")
            cur.execute("DROP TABLE IF EXISTS test_loadbalance")
            cur.execute("CREATE TABLE test_loadbalance (n varchar)")
            # record which node the client has connected to
            for i in range(rowsToInsert):
                with self._connect() as conn1:
                    cur1 = conn1.cursor()
                    cur1.execute("INSERT INTO test_loadbalance ("
                                 "SELECT node_name FROM sessions "
                                 "WHERE session_id = (SELECT current_session()))")
                    conn1.commit()

            cur.execute("SELECT count(n)=3 FROM test_loadbalance GROUP BY n")
            res = cur.fetchall()
            # verify that all db_node_num nodes are represented equally
            self.assertEqual(len(res), self.db_node_num)
            for i in res:
                self.assertEqual(i, [True])


exec(LoadBalanceTestCase.createPrepStmtClass())



# Copyright (c) 2020 Micro Focus or one of its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
CopyError message

In the copy-local protocol, the frontend can terminate the cycle by sending a
CopyError message, which will cause the COPY SQL statement to fail with an error.
"""

from __future__ import print_function, division, absolute_import

from struct import pack

from ..message import BulkFrontendMessage


class CopyError(BulkFrontendMessage):
    message_id = b'e'

    def __init__(self, error_msg, stack_trace=None):
        BulkFrontendMessage.__init__(self)
        self.error_msg = error_msg.encode('utf-8')
        self.file_name = stack_trace[0].encode('utf-8') if stack_trace else b''
        self.line_number = stack_trace[1] if stack_trace else 0
        self.func_name = stack_trace[2].encode('utf-8') if stack_trace else b''

    def read_bytes(self):
        bytes_ = pack('!{0}sxI{1}sx{2}sx'.format(
                      len(self.file_name), len(self.func_name), len(self.error_msg)),
                      self.file_name, self.line_number, self.func_name, self.error_msg)
        return bytes_



# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, division, absolute_import

from .vertica.connection import Connection, connect, parse_dsn

# Importing exceptions for compatibility with dbapi 2.0.
# See: PEP 249 - Python Database API 2.0
#      https://www.python.org/dev/peps/pep-0249/#exceptions
from . import errors
from .errors import (
    Error, Warning, DataError, DatabaseError, IntegrityError, InterfaceError,
    InternalError, NotSupportedError, OperationalError, ProgrammingError)

# Main module for this library.
__author__ = 'Vertica'
__copyright__ = 'Copyright (c) 2018-2020 Micro Focus or one of its affiliates.'
__license__ = 'Apache 2.0'

__all__ = ['Connection', 'PROTOCOL_VERSION', 'version_info', 'apilevel', 'threadsafety',
           'paramstyle', 'connect', 'parse_dsn', 'Error', 'Warning', 'DataError', 'DatabaseError',
           'IntegrityError', 'InterfaceError', 'InternalError', 'NotSupportedError',
           'OperationalError', 'ProgrammingError']

# The version number of this library.
version_info = (1, 0, 1)
__version__ = '.'.join(map(str, version_info))

# The protocol version (3.8) implemented in this library.
PROTOCOL_VERSION = 3 << 16 | 8

apilevel = 2.0
threadsafety = 1  # Threads may share the module, but not connections!

# Accepted paramstyles are
#   'qmark' = Question mark style, e.g. '...WHERE name=?'
#   'named' = Named style, e.g. '...WHERE name=:name'
#   'format' = ANSI C printf format codes, e.g. '...WHERE name=%s'
paramstyle = 'named'



# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, division, absolute_import

from collections import namedtuple
from datetime import date

from .base import VerticaPythonIntegrationTestCase
from ... import errors

DateTestingCase = namedtuple("DateTestingCase", ["string", "template", "date"])


class DateParsingTestCase(VerticaPythonIntegrationTestCase):
    """Testing DATE type parsing with focus on 'AD'/'BC'.

    Note: the 'BC' or 'AD' era indicators in Vertica's date format seem to make Vertica behave as
    follows:
        1. Both 'BC' and 'AD' are simply a flags that tell Vertica: include era indicator if the
        date is Before Christ
        2. Dates in 'AD' will never include era indicator
    """

    def _test_dates(self, test_cases, msg=None):
        with self._connect() as conn:
            cur = conn.cursor()
            for tc in test_cases:
                cur.execute("SELECT TO_DATE('{0}', '{1}')".format(tc.string, tc.template))
                res = cur.fetchall()
                self.assertListOfListsEqual(res, [[tc.date]], msg=msg)

    def _test_not_supported(self, test_cases, msg=None):
        with self._connect() as conn:
            cur = conn.cursor()
            for tc in test_cases:
                with self.assertRaises(errors.NotSupportedError, msg=msg):
                    cur.execute("SELECT TO_DATE('{0}', '{1}')".format(tc.string, tc.template))
                    res = cur.fetchall()
                    self.assertListOfListsEqual(res, [[tc.date]])

    def test_no_to_no(self):
        test_cases = [
            DateTestingCase('1985-10-25', 'YYYY-MM-DD', date(1985, 10, 25)),
            DateTestingCase('1955-11-12', 'YYYY-MM-DD', date(1955, 11, 12)),
            DateTestingCase('1885-01-01', 'YYYY-MM-DD', date(1885, 1, 1)),
            DateTestingCase('2015-10-21', 'YYYY-MM-DD', date(2015, 10, 21)),
        ]
        self._test_dates(test_cases=test_cases, msg='no indicator -> no indicator')

    def test_ad_to_no(self):
        test_cases = [
            DateTestingCase('1985-10-25 AD', 'YYYY-MM-DD', date(1985, 10, 25)),
            DateTestingCase('1955-11-12 AD', 'YYYY-MM-DD', date(1955, 11, 12)),
            DateTestingCase('1885-01-01 AD', 'YYYY-MM-DD', date(1885, 1, 1)),
            DateTestingCase('2015-10-21 AD', 'YYYY-MM-DD', date(2015, 10, 21)),
        ]
        self._test_dates(test_cases=test_cases, msg='AD indicator -> no indicator')

    def test_bc_to_no(self):
        test_cases = [
            DateTestingCase('1985-10-25 BC', 'YYYY-MM-DD', date(1985, 10, 25)),
            DateTestingCase('1955-11-12 BC', 'YYYY-MM-DD', date(1955, 11, 12)),
            DateTestingCase('1885-01-01 BC', 'YYYY-MM-DD', date(1885, 1, 1)),
            DateTestingCase('2015-10-21 BC', 'YYYY-MM-DD', date(2015, 10, 21)),
        ]
        self._test_dates(test_cases=test_cases, msg='BC indicator -> no indicator')

    def test_no_to_ad(self):
        test_cases = [
            DateTestingCase('1985-10-25', 'YYYY-MM-DD AD', date(1985, 10, 25)),
            DateTestingCase('1955-11-12', 'YYYY-MM-DD AD', date(1955, 11, 12)),
            DateTestingCase('1885-01-01', 'YYYY-MM-DD AD', date(1885, 1, 1)),
            DateTestingCase('2015-10-21', 'YYYY-MM-DD AD', date(2015, 10, 21)),
        ]
        self._test_dates(test_cases=test_cases, msg='no indicator -> AD indicator')

    def test_ad_to_ad(self):
        test_cases = [
            DateTestingCase('1985-10-25 AD', 'YYYY-MM-DD AD', date(1985, 10, 25)),
            DateTestingCase('1955-11-12 AD', 'YYYY-MM-DD AD', date(1955, 11, 12)),
            DateTestingCase('1885-01-01 AD', 'YYYY-MM-DD AD', date(1885, 1, 1)),
            DateTestingCase('2015-10-21 AD', 'YYYY-MM-DD AD', date(2015, 10, 21)),
        ]
        self._test_dates(test_cases=test_cases, msg='AD indicator -> AD indicator')

    def test_bc_to_ad(self):
        test_cases = [
            DateTestingCase('1985-10-25 BC', 'YYYY-MM-DD AD', date(1985, 10, 25)),
            DateTestingCase('1955-11-12 BC', 'YYYY-MM-DD AD', date(1955, 11, 12)),
            DateTestingCase('1885-01-01 BC', 'YYYY-MM-DD AD', date(1885, 1, 1)),
            DateTestingCase('2015-10-21 BC', 'YYYY-MM-DD AD', date(2015, 10, 21)),
        ]
        self._test_not_supported(test_cases=test_cases, msg='BC indicator -> AD indicator')

    def test_no_to_bc(self):
        test_cases = [
            DateTestingCase('1985-10-25', 'YYYY-MM-DD BC', date(1985, 10, 25)),
            DateTestingCase('1955-11-12', 'YYYY-MM-DD BC', date(1955, 11, 12)),
            DateTestingCase('1885-01-01', 'YYYY-MM-DD BC', date(1885, 1, 1)),
            DateTestingCase('2015-10-21', 'YYYY-MM-DD BC', date(2015, 10, 21)),
        ]
        self._test_dates(test_cases=test_cases, msg='no indicator -> BC indicator')

    def test_ad_to_bc(self):
        test_cases = [
            DateTestingCase('1985-10-25 AD', 'YYYY-MM-DD BC', date(1985, 10, 25)),
            DateTestingCase('1955-11-12 AD', 'YYYY-MM-DD BC', date(1955, 11, 12)),
            DateTestingCase('1885-01-01 AD', 'YYYY-MM-DD BC', date(1885, 1, 1)),
            DateTestingCase('2015-10-21 AD', 'YYYY-MM-DD BC', date(2015, 10, 21)),
        ]
        self._test_dates(test_cases=test_cases, msg='AD indicator -> BC indicator')

    def test_bc_to_bc(self):
        test_cases = [
            DateTestingCase('1985-10-25 BC', 'YYYY-MM-DD BC', date(1985, 10, 25)),
            DateTestingCase('1955-11-12 BC', 'YYYY-MM-DD BC', date(1955, 11, 12)),
            DateTestingCase('1885-01-01 BC', 'YYYY-MM-DD BC', date(1885, 1, 1)),
            DateTestingCase('2015-10-21 BC', 'YYYY-MM-DD BC', date(2015, 10, 21)),
        ]
        self._test_not_supported(test_cases=test_cases, msg='BC indicator -> BC indicator')


exec(DateParsingTestCase.createPrepStmtClass())



# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, division, absolute_import

from .base import VerticaPythonIntegrationTestCase


class UnicodeTestCase(VerticaPythonIntegrationTestCase):
    def test_unicode_query(self):
        value = u'\u16a0'
        query = u"SELECT '{0}'".format(value)

        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute(query)
            res = cur.fetchone()

        self.assertResultEqual(value, res[0])

    def test_unicode_list_parameter(self):
        values = [u'\u00f1', 'foo', 3]
        query = u"SELECT {0}".format(", ".join(["%s"] * len(values)))

        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute(query, tuple(values))
            results = cur.fetchone()

        for val, res in zip(values, results):
            self.assertResultEqual(val, res)

    def test_unicode_named_parameter_binding(self):
        values = [u'\u16b1', 'foo', 3]
        keys = [u'\u16a0', 'foo', 3]

        query = u"SELECT {0}".format(", ".join([u":{0}".format(key) for key in keys]))

        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute(query, dict(zip(keys, values)))
            results = cur.fetchone()

        for val, res in zip(values, results):
            self.assertResultEqual(val, res)

    def test_string_query(self):
        value = u'test'
        query = u"SELECT '{0}'".format(value)

        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute(query)
            res = cur.fetchone()

        self.assertEqual(value, res[0])

    def test_string_named_parameter_binding(self):
        key = u'test'
        value = u'value'
        query = u"SELECT :{0}".format(key)

        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute(query, {key: value})
            res = cur.fetchone()

        self.assertResultEqual(value, res[0])

    # unit test for issue #160
    def test_null_named_parameter_binding(self):
        key = u'test'
        value = None
        query = u"SELECT :{0}".format(key)

        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute(query, {key: value})
            res = cur.fetchone()

        self.assertResultEqual(value, res[0])

    # unit test for issue #160
    def test_null_list_parameter(self):
        values = [u'\u00f1', 'foo', None]
        query = u"SELECT {0}".format(", ".join(["%s"] * len(values)))

        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute(query, tuple(values))
            results = cur.fetchone()

        for val, res in zip(values, results):
            self.assertResultEqual(val, res)



# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, division, absolute_import

import pytest

from ... import errors, connect
from ..common.base import VerticaPythonTestCase


@pytest.mark.integration_tests
class VerticaPythonIntegrationTestCase(VerticaPythonTestCase):
    """
    Base class for tests that connect to a Vertica database to run stuffs.

    This class is responsible for managing the environment variables and
    connection info used for all the tests, and provides support code
    to do common assertions and execute common queries.
    """

    @classmethod
    def setUpClass(cls):
        config_list = ['log_dir', 'log_level', 'host', 'port',
                       'user', 'password', 'database']
        cls.test_config = cls._load_test_config(config_list)

        # Test logger
        logfile = cls._setup_logger('integration_tests',
                      cls.test_config['log_dir'], cls.test_config['log_level'])

        # Connection info
        # Note: The server-side prepared statements is disabled here. Please
        #       see cls.createPrepStmtClass() below.
        cls._conn_info = {
            'host': cls.test_config['host'],
            'port': cls.test_config['port'],
            'database': cls.test_config['database'],
            'user': cls.test_config['user'],
            'password': cls.test_config['password'],
            'log_level': cls.test_config['log_level'],
            'log_path': logfile,
        }
        cls.db_node_num = cls._get_node_num()
        cls.logger.info("Number of database node(s) = {}".format(cls.db_node_num))

    @classmethod
    def tearDownClass(cls):
        pass

    @classmethod
    def _connect(cls):
        """Connects to vertica.

        :return: a connection to vertica.
        """
        return connect(**cls._conn_info)

    @classmethod
    def _get_node_num(cls):
        """Executes a query to get the number of nodes in the database

        :return: the number of database nodes
        """
        with cls._connect() as conn:
            cur = conn.cursor()
            cur.execute("SELECT count(*) FROM nodes WHERE node_state='UP'")
            return cur.fetchone()[0]

    @classmethod
    def createPrepStmtClass(cls):
        """Generates the code of a new subclass that has the same tests as this
        class but turns on the server-side prepared statements. To ensure test
        coverage, this method should be used if tests are not sensitive to
        paramstyles (or query protocols).
        Usage: "exec(xxxTestCase.createPrepStmtClass())"

        :return: a string acceptable by exec() to define the class
        """
        base_cls_name = cls.__name__
        cls_name = 'PrepStmt' + base_cls_name
        code = ('class ' + cls_name + '(' + base_cls_name + '):\n'
                '  @classmethod\n'
                '  def setUpClass(cls):\n'
                '    super(' + cls_name + ', cls).setUpClass()\n'
                "    cls._conn_info['use_prepared_statements'] = True")
        return code

    def _query_and_fetchall(self, query):
        """Creates a new connection, executes a query and fetches all the results.

        :param query: query to execute
        :return: all fetched results as returned by cursor.fetchall()
        """
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute(query)
            results = cur.fetchall()

        return results

    def _query_and_fetchone(self, query):
        """Creates a new connection, executes a query and fetches one result.

        :param query: query to execute
        :return: the first result fetched by cursor.fetchone()
        """
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute(query)
            result = cur.fetchone()

        return result

    # Common assertions
    def assertConnectionFail(self,
        err_type=errors.ConnectionError,
        err_msg='Failed to establish a connection to the primary server or any backup address.'):
        with pytest.raises(err_type, match=err_msg):
            with self._connect() as conn:
                pass

    def assertConnectionSuccess(self):
        try:
            with self._connect() as conn:
                pass
        except Exception as e:
            self.fail('Connection failed: {0}'.format(e))

    # Some tests require server-side setup
    # In that case, tests that depend on that setup should be skipped to prevent false failures
    # Tests that depend on the server-setup should call these methods to express requirements
    def require_DB_nodes_at_least(self, min_node_num):
        if not isinstance(min_node_num, int):
            err_msg = "Node number '{0}' must be an instance of 'int'".format(min_node_num)
            raise TypeError(err_msg)
        if min_node_num <= 0:
            err_msg = "Node number {0} must be a positive integer".format(min_node_num)
            raise ValueError(err_msg)

        if self.db_node_num < min_node_num:
            msg = ("The test requires a database that has at least {0} node(s), "
                   "but this database has only {1} available node(s).").format(
                   min_node_num, self.db_node_num)
            self.skipTest(msg)

    def require_protocol_at_least(self, min_protocol_version):
        with self._connect() as conn:
            effective_protocol = conn.parameters['protocol_version']
        if effective_protocol < min_protocol_version:
            msg = ("The test requires the effective protocol version to be at "
                   "least {}.{}, but the current version is {}.{}.").format(
                   min_protocol_version >> 16, min_protocol_version & 0x0000ffff,
                   effective_protocol >> 16, effective_protocol & 0x0000ffff)
            self.skipTest(msg)



print("import: 'vertica_python'")
import vertica_python

print("import: 'vertica_python.tests'")
import vertica_python.tests

print("import: 'vertica_python.vertica'")
import vertica_python.vertica

print("import: 'vertica_python.vertica.messages'")
import vertica_python.vertica.messages

print("import: 'vertica_python.vertica.messages.backend_messages'")
import vertica_python.vertica.messages.backend_messages

print("import: 'vertica_python.vertica.messages.frontend_messages'")
import vertica_python.vertica.messages.frontend_messages




# Copyright (c) 2019-2020 Micro Focus or one of its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

import mock

from .base import VerticaPythonUnitTestCase
from ...vertica.messages import NoticeResponse
from ...errors import QueryError


class NoticeTestCase(VerticaPythonUnitTestCase):
    SAMPLE_DATA = {b'S': 'FATAL',
                   b'H': 'This is a test hint',
                   b'L': '9999',
                   b'M': 'Failure is on purpose'}

    @mock.patch.object(NoticeResponse, '_unpack_data')
    def test_error_message(self, mock_unpack_data):
        mock_unpack_data.return_value = NoticeTestCase.SAMPLE_DATA

        notice = NoticeResponse(b'ignored-due-to-mock')
        self.assertEqual(
            notice.error_message(),
            'Severity: FATAL, Message: Failure is on purpose, Hint: This is a test hint, Line: 9999'
        )

    @mock.patch.object(NoticeResponse, '_unpack_data')
    def test_attribute_properties(self, mock_unpack_data):
        mock_unpack_data.return_value = NoticeTestCase.SAMPLE_DATA

        notice = NoticeResponse(b'ignored-due-to-mock')
        self.assertEqual(notice.severity, 'FATAL')
        self.assertEqual(notice.hint, 'This is a test hint')
        # yes, line is still a string.
        self.assertEqual(notice.line, '9999')
        self.assertEqual(notice.message, 'Failure is on purpose')
        self.assertIsNone(notice.detail)
        self.assertIsNone(notice.sqlstate)

    @mock.patch.object(NoticeResponse, '_unpack_data')
    def test_labeled_values(self, mock_unpack_data):
        mock_unpack_data.return_value = NoticeTestCase.SAMPLE_DATA

        notice = NoticeResponse(b'ignored-due-to-mock')
        self.assertEqual(notice.values, {
            'Severity': 'FATAL',
            'Hint': 'This is a test hint',
            'Line': '9999',
            'Message': 'Failure is on purpose'})

    @mock.patch.object(NoticeResponse, '_unpack_data')
    def test_query_error(self, mock_unpack_data):
        mock_unpack_data.return_value = NoticeTestCase.SAMPLE_DATA

        notice = NoticeResponse(b'ignored-due-to-mock')
        query_error = QueryError(notice, 'Select Fake();')

        self.assertEqual(query_error.severity, 'FATAL')
        self.assertEqual(query_error.hint, 'This is a test hint')
        self.assertEqual(query_error.line, '9999')
        self.assertEqual(query_error.message, 'Failure is on purpose')
        self.assertIsNone(query_error.detail)
        self.assertIsNone(query_error.sqlstate)

        self.assertEqual(
            str(query_error),
            'Severity: FATAL, Message: Failure is on purpose, Hint: This is a test hint, Line: 9999, SQL: \'Select Fake();\'')



# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import print_function, division, absolute_import

from .base import VerticaPythonUnitTestCase
from ...errors import VerticaSyntaxError
from ...vertica.messages.backend_messages.error_response import ErrorResponse

import pickle

# Using a subclass of ErrorResponse for this test, to avoid the complexity of
# creating an ErrorResponse object. At the time of writing, ErrorResponse and
# other BackendMessage instances can only be created from server-provided data.
#
# This subclass allows for simpler instantiation without binding to any details
# of server data serialization, a la NoticeResponseAttrMixin and NoticeResponse
class MockErrorResponse(ErrorResponse):
    def __init__(self):
        # does NOT call super
        self._notice_attrs = {}

    def error_message(self):
        return "Manufactured error message for testing"

class ErrorsTestCase(VerticaPythonUnitTestCase):
    def test_pickling(self):
        err_response = MockErrorResponse()
        sql = "select 1;"
        exc = VerticaSyntaxError(err_response, sql)

        serde = pickle.loads(pickle.dumps(exc))

        assert isinstance(serde, VerticaSyntaxError)
        assert str(serde) == str(exc)
        assert isinstance(serde.error_response, MockErrorResponse)
        assert serde.error_response.error_message() == err_response.error_message()
        assert serde.sql == exc.sql



# Copyright (c) 2020 Micro Focus or one of its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import print_function, division, absolute_import

from struct import unpack

from ..message import BackendMessage


class LoadFile(BackendMessage):
    message_id = b'H'

    def __init__(self, data):
        BackendMessage.__init__(self)
        unpacked = unpack('!{0}sx'.format(data.find(b'\x00')), data)
        self.filename = unpacked[0].decode('utf-8')

    def __str__(self):
        return "LoadFile: name = {}".format(self.filename)


BackendMessage.register(LoadFile)



# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Startup message

To begin a session, the frontend opens a connection to the backend and sends a
Startup message.
"""

from __future__ import print_function, division, absolute_import

import platform
import os
from struct import pack

# noinspection PyUnresolvedReferences,PyCompatibility

import vertica_python
from ..message import BulkFrontendMessage


class Startup(BulkFrontendMessage):
    message_id = None

    def __init__(self, user, database, session_label, os_user_name):
        BulkFrontendMessage.__init__(self)

        try:
            os_platform = platform.platform()
        except Exception as e:
            os_platform = ''
            print("WARN: Cannot get the OS info: {}".format(str(e)))

        try:
            pid = str(os.getpid())
        except Exception as e:
            pid = '0'
            print("WARN: Cannot get the process ID: {}".format(str(e)))

        self.parameters = {
            b'user': user,
            b'database': database,
            b'client_label': session_label,
            b'client_type': 'vertica-python',
            b'client_version': vertica_python.__version__,
            b'client_os': os_platform,
            b'client_os_user_name': os_user_name,
            b'client_pid': pid,
        }

    def read_bytes(self):
        # The fixed protocol version is followed by pairs of parameter name and value strings.
        # A zero byte is required as a terminator after the last name/value pair.
        # Parameters can appear in any order.
        fixed_protocol_version = 3 << 16 | 5
        bytes_ = pack('!I', fixed_protocol_version)

        # The frontend sends a requested protocol version to the backend.
        # Old servers (protocol < 3.7) ignore this value and use the fixed protocol version.
        # New servers (protocol >= 3.7) would try to find the common protocol
        # version in use for both client and server, and send back a ParameterStatus
        # message (key='protocol_version', value=<effective protocol version>)
        bytes_ += pack('!16sxIx', b'protocol_version', vertica_python.PROTOCOL_VERSION)

        for k in self.parameters:
            v = self.parameters[k].encode('utf-8')
            bytes_ += pack('!{}sx{}sx'.format(len(k), len(v)), k, v)

        bytes_ += pack('x')
        return bytes_



# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, division, absolute_import

from collections import namedtuple
from datetime import datetime

from .base import VerticaPythonUnitTestCase
from ...vertica.column import timestamp_parse

TimestampTestingCase = namedtuple("TimestampTestingCase", ["string", "timestamp"])


class TimestampParsingTestCase(VerticaPythonUnitTestCase):
    def _test_timestamps(self, test_cases, msg=None):
        for tc in test_cases:
            self.assertEqual(timestamp_parse(tc.string), tc.timestamp, msg=msg)

    def test_timestamp_second_resolution(self):
        test_cases = [  # back to the future dates
            TimestampTestingCase(
                '1985-10-26 01:25:01',
                datetime(year=1985, month=10, day=26, hour=1, minute=25, second=1)
            ),
            TimestampTestingCase(
                '1955-11-12 22:55:02',
                datetime(year=1955, month=11, day=12, hour=22, minute=55, second=2)
            ),
            TimestampTestingCase(
                '2015-10-21 11:12:03',
                datetime(year=2015, month=10, day=21, hour=11, minute=12, second=3)
            ),
            TimestampTestingCase(
                '1885-01-01 01:02:04',
                datetime(year=1885, month=1, day=1, hour=1, minute=2, second=4)
            ),
            TimestampTestingCase(
                '1885-09-02 02:03:05',
                datetime(year=1885, month=9, day=2, hour=2, minute=3, second=5)
            ),
        ]
        self._test_timestamps(test_cases=test_cases, msg='timestamp second resolution')

    def test_timestamp_microsecond_resolution(self):
        test_cases = [  # back to the future dates
            TimestampTestingCase(
                '1985-10-26 01:25:01.1',
                datetime(year=1985, month=10, day=26, hour=1, minute=25, second=1,
                         microsecond=100000)
            ),
            TimestampTestingCase(
                '1955-11-12 22:55:02.01',
                datetime(year=1955, month=11, day=12, hour=22, minute=55, second=2,
                         microsecond=10000)
            ),
            TimestampTestingCase(
                '2015-10-21 11:12:03.001',
                datetime(year=2015, month=10, day=21, hour=11, minute=12, second=3,
                         microsecond=1000)
            ),
            TimestampTestingCase(
                '1885-01-01 01:02:04.000001',
                datetime(year=1885, month=1, day=1, hour=1, minute=2, second=4,
                         microsecond=1)
            ),
            TimestampTestingCase(
                '1885-09-02 02:03:05.002343',
                datetime(year=1885, month=9, day=2, hour=2, minute=3, second=5,
                         microsecond=2343)
            ),
        ]
        self._test_timestamps(test_cases=test_cases, msg='timestamp microsecond resolution')

    def test_timestamp_year_over_9999_second_resolution(self):
        # Asserts that years over 9999 are truncated to 9999
        test_cases = [
            TimestampTestingCase(
                '19850-10-26 01:25:01',
                datetime(year=9999, month=10, day=26, hour=1, minute=25, second=1)
            ),
            TimestampTestingCase(
                '10000-11-12 22:55:02',
                datetime(year=9999, month=11, day=12, hour=22, minute=55, second=2)
            ),
            TimestampTestingCase(
                '9999-10-21 11:12:03',
                datetime(year=9999, month=10, day=21, hour=11, minute=12, second=3)
            ),
            TimestampTestingCase(
                '18850-01-01 01:02:04',
                datetime(year=9999, month=1, day=1, hour=1, minute=2, second=4)
            ),
            TimestampTestingCase(
                '18850-09-02 02:03:05',
                datetime(year=9999, month=9, day=2, hour=2, minute=3, second=5)
            ),
        ]
        self._test_timestamps(test_cases=test_cases, msg='timestamp past 9999 second resolution')

    def test_timestamp_year_over_9999_microsecond_resolution(self):
        test_cases = [
            TimestampTestingCase(
                '19850-10-26 01:25:01.1',
                datetime(year=9999, month=10, day=26, hour=1, minute=25, second=1,
                         microsecond=100000)
            ),
            TimestampTestingCase(
                '10000-11-12 22:55:02.01',
                datetime(year=9999, month=11, day=12, hour=22, minute=55, second=2,
                         microsecond=10000)
            ),
            TimestampTestingCase(
                '9999-10-21 11:12:03.001',
                datetime(year=9999, month=10, day=21, hour=11, minute=12, second=3,
                         microsecond=1000)
            ),
            TimestampTestingCase(
                '18850-01-01 01:02:04.000001',
                datetime(year=9999, month=1, day=1, hour=1, minute=2, second=4,
                         microsecond=1)
            ),
            TimestampTestingCase(
                '18850-09-02 02:03:05.002343',
                datetime(year=9999, month=9, day=2, hour=2, minute=3, second=5,
                         microsecond=2343)
            ),
        ]
        self._test_timestamps(test_cases=test_cases,
                              msg='timestamp past 9999 microsecond resolution')



# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from .base import VerticaPythonIntegrationTestCase


class AuthenticationTestCase(VerticaPythonIntegrationTestCase):
    def setUp(self):
        super(AuthenticationTestCase, self).setUp()
        self._user = self._conn_info['user']
        self._password = self._conn_info['password']

    def tearDown(self):
        self._conn_info['user'] = self._user
        self._conn_info['password'] = self._password
        super(AuthenticationTestCase, self).tearDown()

    def test_SHA512(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("DROP USER IF EXISTS sha512_user")
            cur.execute("DROP AUTHENTICATION IF EXISTS testIPv4hostHash CASCADE")
            cur.execute("DROP AUTHENTICATION IF EXISTS testIPv6hostHash CASCADE")
            cur.execute("DROP AUTHENTICATION IF EXISTS testlocalHash CASCADE")
            try:
                cur.execute("CREATE USER sha512_user IDENTIFIED BY 'password'")
                cur.execute("ALTER USER sha512_user SECURITY_ALGORITHM 'SHA512'")
                cur.execute("CREATE AUTHENTICATION testIPv4hostHash METHOD 'hash' HOST '0.0.0.0/0'")
                cur.execute("GRANT AUTHENTICATION testIPv4hostHash TO sha512_user")
                cur.execute("CREATE AUTHENTICATION testIPv6hostHash METHOD 'hash' HOST '::/0'")
                cur.execute("GRANT AUTHENTICATION testIPv6hostHash TO sha512_user")
                cur.execute("CREATE AUTHENTICATION testlocalHash METHOD 'hash' LOCAL")
                cur.execute("GRANT AUTHENTICATION testlocalHash TO sha512_user")

                self._conn_info['user'] = 'sha512_user'
                self._conn_info['password'] = 'password'

                expire_msg = 'The password for user sha512_user has expired'
                self.assertConnectionFail(err_msg=expire_msg)

                # Test SHA512 connection
                cur.execute("ALTER USER sha512_user IDENTIFIED BY 'password'")
                self.assertConnectionSuccess()

                # Switch to MD5 for hashing and storing the user password
                cur.execute("ALTER USER sha512_user SECURITY_ALGORITHM 'MD5'")
                self.assertConnectionFail(err_msg=expire_msg)

                # Test MD5 connection
                cur.execute("ALTER USER sha512_user IDENTIFIED BY 'password'")
                self.assertConnectionSuccess()
            finally:
                # Must clean up authentication methods within this session, no
                # matter whether an exception has occurred or not, otherwise
                # those authentication methods may affect existing users
                cur.execute("DROP USER IF EXISTS sha512_user")
                cur.execute("DROP AUTHENTICATION IF EXISTS testIPv4hostHash CASCADE")
                cur.execute("DROP AUTHENTICATION IF EXISTS testIPv6hostHash CASCADE")
                cur.execute("DROP AUTHENTICATION IF EXISTS testlocalHash CASCADE")

    def test_password_expire(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("DROP USER IF EXISTS pw_expire_user")
            cur.execute("DROP AUTHENTICATION IF EXISTS testIPv4hostHash CASCADE")
            cur.execute("DROP AUTHENTICATION IF EXISTS testIPv6hostHash CASCADE")
            cur.execute("DROP AUTHENTICATION IF EXISTS testlocalHash CASCADE")
            try:
                cur.execute("CREATE USER pw_expire_user IDENTIFIED BY 'password'")
                cur.execute("CREATE AUTHENTICATION testIPv4hostHash METHOD 'hash' HOST '0.0.0.0/0'")
                cur.execute("GRANT AUTHENTICATION testIPv4hostHash TO pw_expire_user")
                cur.execute("CREATE AUTHENTICATION testIPv6hostHash METHOD 'hash' HOST '::/0'")
                cur.execute("GRANT AUTHENTICATION testIPv6hostHash TO pw_expire_user")
                cur.execute("CREATE AUTHENTICATION testlocalHash METHOD 'hash' LOCAL")
                cur.execute("GRANT AUTHENTICATION testlocalHash TO pw_expire_user")

                # Test connection
                self._conn_info['user'] = 'pw_expire_user'
                self._conn_info['password'] = 'password'
                self.assertConnectionSuccess()

                # Expire the user's password immediately
                cur.execute("ALTER USER pw_expire_user PASSWORD EXPIRE")

                # Expect an error message
                self.assertConnectionFail(
                    err_msg='The password for user pw_expire_user has expired')
            finally:
                # Must clean up authentication methods within this session, no
                # matter whether an exception has occurred or not, otherwise
                # those authentication methods may affect existing users
                cur.execute("DROP USER IF EXISTS pw_expire_user")
                cur.execute("DROP AUTHENTICATION IF EXISTS testIPv4hostHash CASCADE")
                cur.execute("DROP AUTHENTICATION IF EXISTS testIPv6hostHash CASCADE")
                cur.execute("DROP AUTHENTICATION IF EXISTS testlocalHash CASCADE")


exec(AuthenticationTestCase.createPrepStmtClass())



# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, division, absolute_import

from datetime import date, datetime, time
from decimal import Decimal
from io import open
from uuid import UUID
import logging
import os
import pytest
import re
import shutil
import sys
import tempfile

from parameterized import parameterized

from .base import VerticaPythonIntegrationTestCase
from ... import errors

"""
There are a couple of testcases in this file, they are
1. CursorTestCase:
        general cursor tests, not sensitive to query protocols.
2. SimpleQueryTestCase:
        simple query protocol tests in execute().
3. SimpleQueryExecutemanyTestCase:
        simple query protocol tests in executemany().
4. PreparedStatementTestCase:
        prepared statements tests in both execute() and executemany().

Different query protocols use different paramstyles:
    - simple query protocol: 'named' and 'format' paramstyles
    - extended query protocol (prepared statements): 'qmark' paramstyle
"""


class CursorTestCase(VerticaPythonIntegrationTestCase):
    def setUp(self):
        super(CursorTestCase, self).setUp()
        self._table = 'cursor_test'
        self._init_table()

    def _init_table(self):
        with self._connect() as conn:
            cur = conn.cursor()
            # clean old table
            cur.execute("DROP TABLE IF EXISTS {0}".format(self._table))

            # create test table
            cur.execute("""CREATE TABLE {0} (
                                a INT,
                                b VARCHAR(32)
                           )
                        """.format(self._table))

    def tearDown(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("DROP TABLE IF EXISTS {0}".format(self._table))
        super(CursorTestCase, self).tearDown()

    def test_multi_inserts_and_transaction(self):
        with self._connect() as conn1, self._connect() as conn2:
            cur1 = conn1.cursor()
            cur2 = conn2.cursor()

            # insert data without a commit
            cur1.execute("INSERT INTO {0} (a, b) VALUES (2, 'bb')".format(self._table))

            # verify we can see it from this cursor
            cur1.execute("SELECT a, b FROM {0} WHERE a = 2".format(self._table))
            res_from_cur_1_before_commit = cur1.fetchall()
            self.assertListOfListsEqual(res_from_cur_1_before_commit, [[2, 'bb']])

            # verify we cant see it from other cursor
            cur2.execute("SELECT a, b FROM {0} WHERE a = 2".format(self._table))

            res_from_cur2_before_commit = cur2.fetchall()
            self.assertListOfListsEqual(res_from_cur2_before_commit, [])

            # insert more data then commit
            cur1.execute("INSERT INTO {0} (a, b) VALUES (3, 'cc')".format(self._table))
            cur1.execute("COMMIT")

            # verify we can see it from this cursor
            cur1.execute(
                "SELECT a, b FROM {0} WHERE a = 2 OR a = 3 ORDER BY a".format(self._table))
            res_from_cur1_after_commit = cur1.fetchall()
            self.assertListOfListsEqual(res_from_cur1_after_commit, [[2, 'bb'], [3, 'cc']])

            # verify we can see it from other cursor
            cur2.execute(
                "SELECT a, b FROM {0} WHERE a = 2 OR a = 3 ORDER BY a".format(self._table))
            res_from_cur2_after_commit = cur2.fetchall()
            self.assertListOfListsEqual(res_from_cur2_after_commit, [[2, 'bb'], [3, 'cc']])

    def test_conn_commit(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("INSERT INTO {0} (a, b) VALUES (5, 'cc')".format(self._table))
            conn.commit()

        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("SELECT a, b FROM {0} WHERE a = 5".format(self._table))
            res = cur.fetchall()

        self.assertListOfListsEqual(res, [[5, 'cc']])

    def test_delete(self):
        with self._connect() as conn:
            cur = conn.cursor()

            cur.execute("INSERT INTO {0} (a, b) VALUES (5, 'cc')".format(self._table))
            self.assertEqual(cur.rowcount, -1)
            update_res = cur.fetchall()
            self.assertListOfListsEqual(update_res, [[1]])
            conn.commit()

            # validate delete count
            cur.execute("DELETE FROM {0} WHERE a = 5".format(self._table))
            self.assertEqual(cur.rowcount, -1)
            delete_res = cur.fetchall()
            self.assertListOfListsEqual(delete_res, [[1]])
            conn.commit()

            # validate deleted
            cur.execute("SELECT a, b FROM {0} WHERE a = 5".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [])

    def test_update(self):
        with self._connect() as conn:
            cur = conn.cursor()

            cur.execute("INSERT INTO {0} (a, b) VALUES (5, 'cc')".format(self._table))
            # validate insert count
            insert_res = cur.fetchall()
            self.assertListOfListsEqual(insert_res, [[1]], msg='Bad INSERT response')
            conn.commit()

            cur.execute("UPDATE {0} SET b = 'ff' WHERE a = 5".format(self._table))
            # validate update count
            assert cur.rowcount == -1
            update_res = cur.fetchall()
            self.assertListOfListsEqual(update_res, [[1]], msg='Bad UPDATE response')
            conn.commit()

            cur.execute("SELECT a, b FROM {0} WHERE a = 5".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[5, 'ff']])

    def test_copy_null(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.copy("COPY {0} (a, b) FROM STDIN DELIMITER ','".format(self._table),
                     "1,\n,foo")
            cur.execute("SELECT a, b FROM {0} ORDER BY a ASC".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[None, 'foo'], [1, None]])

    def test_copy_with_string(self):
        with self._connect() as conn1, self._connect() as conn2:
            cur1 = conn1.cursor()
            cur2 = conn2.cursor()

            cur1.copy("COPY {0} (a, b) FROM STDIN DELIMITER ','".format(self._table),
                      "1,foo\n2,bar")
            # no commit necessary for copy
            cur1.execute("SELECT a, b FROM {0} WHERE a = 1".format(self._table))
            res_from_cur1 = cur1.fetchall()
            self.assertListOfListsEqual(res_from_cur1, [[1, 'foo']])

            cur2.execute("SELECT a, b FROM {0} WHERE a = 2".format(self._table))
            res_from_cur2 = cur2.fetchall()
            self.assertListOfListsEqual(res_from_cur2, [[2, 'bar']])

    # integration test for #345
    def test_copy_with_file_like_object(self):
        class FileWrapper:
            read = property(lambda self: self.file.read)
            seek = property(lambda self: self.file.seek)
            write = property(lambda self: self.file.write)

            def __init__(self, file):
                self.file = file

        with tempfile.TemporaryFile() as f, self._connect() as conn:
            wrapped_f = FileWrapper(f)  # object with a read() method
            wrapped_f.write(b"1,foo\n2,bar")
            # move rw pointer to top of file
            wrapped_f.seek(0)

            cur = conn.cursor()
            cur.copy(
                "COPY {0} (a, b) FROM STDIN DELIMITER ','".format(self._table),
                wrapped_f
            )
            cur.execute("SELECT * FROM {0} ORDER BY a ASC".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[1, 'foo'], [2, 'bar']])

    # integration test for #325
    @parameterized.expand([
        (tempfile.NamedTemporaryFile,),
        (tempfile.SpooledTemporaryFile,),
        (tempfile.TemporaryFile,),
    ])
    def test_copy_with_temporary_file(self, temp_file_type):
        with temp_file_type() as f, self._connect() as conn:
            f.write(b"1,foo\n2,bar")
            f.seek(0)

            cur = conn.cursor()
            cur.copy(
                "COPY {0} (a, b) FROM STDIN DELIMITER ','".format(self._table),
                f,
            )
            cur.execute("SELECT * FROM {0} ORDER BY a ASC".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[1, 'foo'], [2, 'bar']])

    def test_copy_with_file(self):
        with tempfile.TemporaryFile() as f, self._connect() as conn1, self._connect() as conn2:
            f.write(b"1,foo\n2,bar")
            # move rw pointer to top of file
            f.seek(0)

            cur1 = conn1.cursor()
            cur2 = conn2.cursor()

            cur1.copy("COPY {0} (a, b) FROM STDIN DELIMITER ','".format(self._table),
                      f)
            # no commit necessary for copy
            cur1.execute("SELECT a, b FROM {0} WHERE a = 1".format(self._table))
            res_from_cur1 = cur1.fetchall()
            self.assertListOfListsEqual(res_from_cur1, [[1, 'foo']])

            cur2.execute("SELECT a, b FROM {0} WHERE a = 2".format(self._table))
            res_from_cur2 = cur2.fetchall()
            self.assertListOfListsEqual(res_from_cur2, [[2, 'bar']])

    def test_copy_with_closed_file(self):
        with tempfile.TemporaryFile() as f, self._connect() as conn:
            f.write(b"1,foo\n2,bar")
            # move rw pointer to top of file
            f.seek(0)

            cur = conn.cursor()
            f.close()
            with pytest.raises(errors.DataError, match='closed file'):
                cur.copy("COPY {0} (a, b) FROM STDIN DELIMITER ','".format(self._table),
                          f)
            # Must not close the cursor object and able to successfully run queries
            cur.execute("SELECT 1;")
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[1]])

    # unit test for #78
    def test_copy_with_data_in_buffer(self):
        with self._connect() as conn:
            cur = conn.cursor()

            cur.execute("SELECT 1;")
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[1]])

            cur.copy("COPY {0} (a, b) FROM STDIN DELIMITER ','".format(self._table),
                     "1,foo\n2,bar")

            cur.execute("SELECT 1;")
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[1]])

    # unit test for #213
    def test_cmd_after_invalid_copy_stmt(self):
        with self._connect() as conn:
            cur = conn.cursor()

            cur.execute("SELECT 1;")
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[1]])

            res = [[]]
            try:
                cur.copy("COPY non_existing_tab(a, b) FROM STDIN DELIMITER ','", "FAIL")
            except errors.MissingRelation as e:
                cur.execute("SELECT 1;")
                res = cur.fetchall()

            self.assertListOfListsEqual(res, [[1]])

    # unit test for #213
    def test_cmd_after_rejected_copy_data(self):
        with self._connect() as conn:
            cur = conn.cursor()

            cur.execute("SELECT 1;")
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[1]])

            res = [[]]
            try:
                cur.copy("COPY {0} (a, b) FROM STDIN DELIMITER ',' ABORT ON ERROR".format(self._table),
                         "FAIL")
            except errors.CopyRejected as e:
                cur.execute("SELECT 1;")
                res = cur.fetchall()

            self.assertListOfListsEqual(res, [[1]])

    def test_with_conn(self):
        with self._connect() as conn:
            cur = conn.cursor()

            cur.execute("INSERT INTO {0} (a, b) VALUES (1, 'aa');".format(self._table))
            conn.commit()
            cur.execute("SELECT a, b FROM {0} WHERE a = 1".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[1, 'aa']])

    def test_iterator(self):
        with self._connect() as conn:
            cur = conn.cursor()
            values = [[1, 'aa'], [2, 'bb'], [3, 'cc']]

            for n, s in values:
                cur.execute("INSERT INTO {} (a, b) VALUES ({}, '{}')"
                            .format(self._table, n, s))
            conn.commit()

            cur.execute("SELECT a, b FROM {0} ORDER BY a ASC".format(self._table))

            for val, res in zip(sorted(values), cur.iterate()):
                self.assertListEqual(res, val)

            remaining = cur.fetchall()
            self.assertListOfListsEqual(remaining, [])

    def test_mid_iterator_execution(self):
        with self._connect() as conn:
            cur = conn.cursor()
            values = [[1, 'aa'], [2, 'bb'], [3, 'cc']]

            for n, s in values:
                cur.execute("INSERT INTO {} (a, b) VALUES ({}, '{}')"
                            .format(self._table, n, s))
            conn.commit()

            cur.execute("SELECT a, b FROM {0} ORDER BY a ASC".format(self._table))

            for val, res in zip(sorted(values), cur.iterate()):
                self.assertListEqual(res, val)
                break  # stop after one comparison

            # make new query and verify result
            cur.execute("SELECT COUNT(*) FROM {0}".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[3]])

    def test_query_errors(self):
        with self._connect() as conn:
            cur = conn.cursor()

            # create table syntax error
            with self.assertRaises(errors.VerticaSyntaxError):
                cur.execute("""CREATE TABLE {0}_fail (
                                a INT,
                                b VARCHAR(32),,,
                                );
                            """.format(self._table))

            # select table not found error
            cur.execute("INSERT INTO {0} (a, b) VALUES (1, 'aa');".format(self._table))
            cur.execute("COMMIT;")
            with self.assertRaises(errors.QueryError):
                cur.execute("SELECT * FROM {0}_fail".format(self._table))

            # generate a user-defined error message
            err_msg = 'USER GENERATED ERROR: test error'
            with pytest.raises(errors.QueryError, match=err_msg):
                cur.execute("SELECT THROW_ERROR('test error')")

            # verify cursor still usable after errors
            cur.execute("SELECT a, b FROM {0} WHERE a = 1".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[1, 'aa']])

    def test_cursor_close_and_reuse(self):
        with self._connect() as conn:
            cur = conn.cursor()

            # insert data
            cur.execute("INSERT INTO {0} (a, b) VALUES (2, 'bb');".format(self._table))
            cur.execute("COMMIT;")

            # (query -> close -> reopen) * 3 times
            for _ in range(3):
                cur.execute("SELECT a, b FROM {0} WHERE a = 2".format(self._table))
                res = cur.fetchall()
                self.assertListOfListsEqual(res, [[2, 'bb']])

                # close and reopen cursor
                cur.close()
                with pytest.raises(errors.InterfaceError, match='Cursor is closed'):
                    cur.execute("SELECT 1;")
                cur = conn.cursor()

    def test_format_quote_unicode(self):
        with self._connect() as conn:
            cur = conn.cursor()
            bad_word = u'Fr\xfchst\xfcck'
            formatted_word = u''.join((u'"', re.escape(bad_word), u'"'))
            self.assertEqual(formatted_word, cur.format_quote(bad_word, True))

    def test_udtype(self):
        poly = "POLYGON ((1 2, 2 3, 3 1, 1 2))"
        line = "LINESTRING (42.1 71, 41.4 70, 41.3 72.9, 42.99 71.46, 44.47 73.21)"
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("DROP TABLE IF EXISTS {0}".format(self._table))
            cur.execute("CREATE TABLE {} (c1 GEOMETRY(10000), c2 GEOGRAPHY(1000))"
                .format(self._table))
            cur.execute("INSERT INTO {} VALUES (ST_GeomFromText('{}'), ST_GeographyFromText('{}'))"
                .format(self._table, poly, line))
            conn.commit()
            cur.execute("SELECT c1, c2, ST_AsText(c1), ST_AsText(c2) FROM {}".format(self._table))

            res = cur.fetchall()
            self.assertEqual(res[0][2], poly)
            self.assertEqual(res[0][3], line)

            datatype_names = [col.type_name for col in cur.description]
            expected = ['geometry', 'geography', 'Long Varchar', 'Long Varchar']
            self.assertListEqual(datatype_names, expected)

            self.assertEqual(cur.description[0].display_size, 10000)
            self.assertEqual(cur.description[1].display_size, 1000)


exec(CursorTestCase.createPrepStmtClass())


class SimpleQueryTestCase(VerticaPythonIntegrationTestCase):
    @classmethod
    def setUpClass(cls):
        super(SimpleQueryTestCase, cls).setUpClass()
        cls._conn_info['use_prepared_statements'] = False
        # Create data files for COPY LOCAL tests
        with tempfile.NamedTemporaryFile(delete=False) as cls._f1:
            cls._f1.write(b"1,foo\n2,bar\nx\xc3\xb1,bla")
        with tempfile.NamedTemporaryFile(delete=False) as cls._f2:
            cls._f2.write(b"4,\n5," + b'a'*10 + b"\n,baz")
        with tempfile.NamedTemporaryFile(delete=False) as cls._f3:
            cls._f3.write(b"10," + b'k'*12 + b"\n11,qux\nxx,corge")
        with tempfile.NamedTemporaryFile(delete=False) as cls._f4:
            cls._f4.write(b"13,flob\nf,quux\n15,xyz")

    @classmethod
    def tearDownClass(cls):
        for f in (cls._f1, cls._f2, cls._f3, cls._f4):
            os.remove(f.name)

    def setUp(self):
        super(SimpleQueryTestCase, self).setUp()
        self._table = 'simplequery_test'
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("DROP TABLE IF EXISTS {0}".format(self._table))

    def tearDown(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("DROP TABLE IF EXISTS {0}".format(self._table))
        super(SimpleQueryTestCase, self).tearDown()

    def test_inline_commit(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {0} (a INT, b VARCHAR)".format(self._table))
            cur.execute("INSERT INTO {0} (a, b) VALUES (1, 'aa'); COMMIT;"
                        .format(self._table))
            cur.execute("SELECT a, b FROM {0} WHERE a = 1".format(self._table))

            # unknown rowcount
            self.assertEqual(cur.rowcount, -1)

            res = cur.fetchall()
            self.assertEqual(cur.rowcount, 1)

            self.assertListOfListsEqual(res, [[1, 'aa']])

    # unit test for #144
    def test_empty_query(self):
        with self._connect() as conn:
            cur = conn.cursor()

            cur.execute("")
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [])

            cur.execute("--select 1")
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [])

            cur.execute("""
                        /*
                        Test block comment
                        */
                        """)
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [])

    # unit test for #175
    def test_datetime_types(self):
        with self._connect() as conn:
            cur = conn.cursor()

            # create test table
            cur.execute("""CREATE TABLE {0} (
                                a INT,
                                b VARCHAR(32),
                                c TIMESTAMP,
                                d DATE,
                                e TIME
                           )
                        """.format(self._table))

            cur.execute("INSERT INTO {0} VALUES (:n, :s, :dt, :d, :t)".format(self._table),
                            {'n': 10, 's': 'aa',
                             'dt': datetime(2018, 9, 7, 15, 38, 19, 769000),
                             'd': date(2018, 9, 7),
                             't': time(13, 50, 9)})
            conn.commit()

            cur.execute("SELECT a, b, c, d, e FROM {0}".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[10, 'aa',
                                               datetime(2018, 9, 7, 15, 38, 19, 769000),
                                               date(2018, 9, 7), time(13, 50, 9)]])

    def test_binary_types(self):
        with self._connect() as conn:
            cur = conn.cursor()

            # create test table
            cur.execute("""CREATE TABLE {0} (
                                a binary(1),
                                b binary(3),
                                c varbinary
                            )
                        """.format(self._table))

            cur.execute("INSERT INTO {0} VALUES (:b, :s1, :s2)".format(self._table),
                        {'b': b'x', 's1': b'xyz', 's2': 'abcde'})
            conn.commit()

            cur.execute("SELECT a, b, c FROM {0}".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[b'x', b'xyz', b'abcde']])

    def test_uuid_type(self):
        with self._connect() as conn:
            cur = conn.cursor()
            x = UUID('{00010203-0405-0607-0809-0a0b0c0d0e0f}')
            y = UUID('00000020-0000-0000-0000-100000000000')
            z = UUID('00100010000000000000000000000000')

            # create test table
            cur.execute("CREATE TABLE {0} ( a uuid, b uuid, c uuid )".format(self._table))
            cur.execute("INSERT INTO {0} VALUES (:u1, :u2, :u3)".format(self._table),
                        {'u1': x, 'u2': y, 'u3': z})
            conn.commit()

            cur.execute("SELECT a, b, c FROM {0}".format(self._table))
            res = cur.fetchall()[0]
            self.assertListEqual([str(i) for i in res], [str(x), str(y), str(z)])

    # unit test for #74
    def test_nextset(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("SELECT 1; SELECT 'foo';")

            res1 = cur.fetchall()
            self.assertEqual(cur.rowcount, 1)
            self.assertListOfListsEqual(res1, [[1]])
            self.assertIsNone(cur.fetchone())
            self.assertTrue(cur.nextset())

            res2 = cur.fetchall()
            self.assertEqual(cur.rowcount, 1)
            self.assertListOfListsEqual(res2, [['foo']])
            self.assertIsNone(cur.fetchone())
            self.assertFalse(cur.nextset())

    # unit test for #74
    def test_nextset_with_delete(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {0} (a INT, b VARCHAR(32))".format(self._table))
            # insert data
            cur.execute("INSERT INTO {0} (a, b) VALUES (1, 'aa')".format(self._table))
            cur.execute("INSERT INTO {0} (a, b) VALUES (2, 'bb')".format(self._table))
            conn.commit()

            cur.execute("""
                          SELECT * FROM {0} ORDER BY a ASC;
                          DELETE FROM {0};
                          SELECT * FROM {0} ORDER BY a ASC;
                        """.format(self._table))

            # check first select results
            res1 = cur.fetchall()
            self.assertListOfListsEqual(res1, [[1, 'aa'], [2, 'bb']])
            self.assertIsNone(cur.fetchone())
            self.assertTrue(cur.nextset())

            # check delete results
            res2 = cur.fetchall()
            self.assertListOfListsEqual(res2, [[2]])
            self.assertIsNone(cur.fetchone())
            self.assertTrue(cur.nextset())

            # check second select results
            res3 = cur.fetchall()
            self.assertListOfListsEqual(res3, [])
            self.assertIsNone(cur.fetchone())
            self.assertFalse(cur.nextset())

    # unit test for #124
    def test_nextset_with_error(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("SELECT 1; SELECT a; SELECT 2")

            # verify data from first query
            res1 = cur.fetchall()
            self.assertListOfListsEqual(res1, [[1]])
            self.assertIsNone(cur.fetchone())

            # second statement results in a query error
            with self.assertRaises(errors.MissingColumn):
                cur.nextset()

    def test_qmark_paramstyle(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {0} (a INT, b VARCHAR)".format(self._table))
            err_msg = 'not all arguments converted during string formatting'
            values = [1, 'aa']
            with pytest.raises(TypeError, match=err_msg):
                cur.execute("INSERT INTO {} VALUES (?, ?)".format(self._table), values)

            cur.execute("INSERT INTO {} VALUES (?, ?)".format(self._table),
                        values, use_prepared_statements=True)
            cur.execute("SELECT * FROM {}".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [values])

    def test_execute_parameters(self):
        with self._connect() as conn:
            cur = conn.cursor()
            all_chars = u"".join(chr(i) for i in range(1, 128))
            backslash_data = u"\\backslash\\ \\data\\\\"
            cur.execute("SELECT :a, :b", parameters={"a": all_chars, "b": backslash_data})
            self.assertEqual([all_chars, backslash_data], cur.fetchone())

    def test_execute_percent_parameters(self):
        with self._connect() as conn:
            cur = conn.cursor()
            all_chars = u"".join(chr(i) for i in range(1, 128))
            backslash_data = u"\\backslash\\ \\data\\\\"
            cur.execute("SELECT %s, %s", parameters=[all_chars, backslash_data])
            self.assertEqual([all_chars, backslash_data], cur.fetchone())

    def test_disabled_copy_local(self):
        self._conn_info['disable_copy_local'] = True

        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {0} (a INT, b VARCHAR(9))".format(self._table))
            with pytest.raises(errors.InterfaceError, match='disabled'):
                cur.execute(
                    "COPY {} FROM LOCAL '{}','{}' DELIMITER ',' ENFORCELENGTH;"
                    "SELECT 100;"
                    .format(self._table, self._f1.name, self._f2.name)
                )
            self.assertListOfListsEqual(cur.fetchall(), [])
            self.assertFalse(cur.nextset())

            # Must not close the cursor object and able to successfully run queries
            cur.execute("SELECT 1;")
            self.assertListOfListsEqual(cur.fetchall(), [[1]])

    def test_copy_local_stdin_input_options(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {0} (a INT, b VARCHAR(9))".format(self._table))
            # No STDIN input
            with pytest.raises(ValueError, match='No STDIN source'):
                cur.execute(
                    "COPY {} FROM LOCAL STDIN DELIMITER ',' ENFORCELENGTH"
                    .format(self._table))
            with pytest.raises(ValueError, match='No STDIN source'):
                cur.execute(
                    "COPY {} FROM LOCAL STDIN DELIMITER ',' ENFORCELENGTH"
                    .format(self._table), copy_stdin=[])
            # Invalid STDIN input
            with pytest.raises(TypeError, match='file-like object'):
                cur.execute(
                    "COPY {} FROM LOCAL STDIN DELIMITER ',' ENFORCELENGTH"
                    .format(self._table), copy_stdin='Not file-like')
            with pytest.raises(TypeError, match='file-like object'):
                cur.execute(
                    "COPY {} FROM LOCAL STDIN DELIMITER ',' ENFORCELENGTH"
                    .format(self._table), copy_stdin=['Not file-like'])
            # A file-like object as STDIN input
            fs = open(self._f1.name)
            cur.execute(
                "COPY {} FROM LOCAL STDIN DELIMITER ',' ENFORCELENGTH"
                .format(self._table), copy_stdin=fs)
            res1 = cur.fetchall()
            self.assertListOfListsEqual(res1, [[2]])
            fs.close()

    def test_copy_local_stdin_multistat(self):
        # Define paths to rejected files
        tmpdir = os.path.dirname(self._f1.name)
        rejdir = os.path.join(tmpdir, 'copylocal1')
        rej1 = os.path.join(rejdir, 'copy_rej.txt')
        except1 = os.path.join(rejdir, 'copy_exception.txt')
        if os.path.isdir(rejdir):
            shutil.rmtree(rejdir)
        rej2 = os.path.join(tmpdir, 'copy_rej2.txt')
        except2 = os.path.join(tmpdir, 'copy_exception2.txt')
        for f in (rej2, except2):
            if os.path.isfile(f):
                os.remove(f)

        # Execute the COPY LOCAL statement as the first and later statement
        # within a query containing multiple statements
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {0} (a INT, b VARCHAR(9))".format(self._table))
            # Feed STDIN
            f1 = open(self._f1.name)
            f2 = open(self._f3.name)
            cur.execute(
                "COPY {} FROM LOCAL STDIN DELIMITER ',' ENFORCELENGTH"
                " REJECTED DATA '{}' EXCEPTIONS '{}';"
                "SELECT 100;"
                "COPY {} FROM LOCAL STDIN DELIMITER ',' ENFORCELENGTH"
                " REJECTED DATA '{}' EXCEPTIONS '{}';"
                "SELECT 200;"
                .format(self._table, rej1, except1,
                        self._table, rej2, except2),
                copy_stdin=[f1, f2]
            )

            res1 = cur.fetchall()
            self.assertListOfListsEqual(res1, [[2]])
            self.assertTrue(cur.nextset())
            f1.close()

            res2 = cur.fetchall()
            self.assertListOfListsEqual(res2, [[100]])
            self.assertTrue(cur.nextset())

            res3 = cur.fetchall()
            self.assertListOfListsEqual(res3, [[1]])
            self.assertTrue(cur.nextset())
            f2.close()

            res4 = cur.fetchall()
            self.assertListOfListsEqual(res4, [[200]])
            self.assertFalse(cur.nextset())

            # There should be no hang/error in the next execution call
            cur.execute("SELECT * FROM {0} ORDER BY a ASC".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[1, 'foo'], [2, 'bar'], [11, 'qux']])

        # Check rejected files
        with open(rej1, 'r', encoding='utf-8') as f:
            self.assertEqual(f.read(), u'x\u00f1,bla\n')
        with open(except1, 'r', encoding='utf-8') as f:
            content = f.read()
            self.assertTrue(u"Invalid integer format 'x\u00f1' for column 1 (a)" in content)
        with open(rej2, 'r', encoding='utf-8') as f:
            self.assertEqual(f.read(), u'10,kkkkkkkkkkkk\nxx,corge\n')
        with open(except2, 'r', encoding='utf-8') as f:
            content = f.read()
            self.assertTrue(u"The 12-byte value is too long for type Varchar(9), column 2 (b)" in content)
            self.assertTrue(u"Invalid integer format 'xx' for column 1 (a)" in content)

        # Delete data files
        try:
            for f in (rej2, except2):
                os.remove(f)
            shutil.rmtree(rejdir)
        except Exception:
            pass

    def test_copy_local_file_not_exist(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {0} (a INT, b VARCHAR(9))".format(self._table))
            with pytest.raises(OSError, match='not_exist.file does not exist'):
                cur.execute(
                    "COPY {} FROM LOCAL '{}','not_exist.file' DELIMITER ',' ENFORCELENGTH"
                    .format(self._table, self._f1.name))
            self.assertListOfListsEqual(cur.fetchall(), [])
            self.assertFalse(cur.nextset())

            # Must not close the cursor object and able to successfully run queries
            cur.execute("SELECT 1;")
            self.assertListOfListsEqual(cur.fetchall(), [[1]])

    def test_copy_local_glob(self):
        suffix = ".copy_glob_test"
        files = (self._f1.name, self._f2.name, self._f3.name, self._f4.name)
        fdir = os.path.dirname(self._f1.name)
        for f in files:
            shutil.copy(f, f + suffix)
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {0} (a INT, b VARCHAR(9))".format(self._table))
            cur.execute(
                "COPY {} FROM LOCAL '{}' DELIMITER ',' ENFORCELENGTH"
                .format(self._table, os.path.join(fdir, '*' + suffix)))
            self.assertListOfListsEqual(cur.fetchall(), [[7]])
        for f in files:
            os.remove(f + suffix)

    @parameterized.expand([(True,), (False,)])
    def test_copy_local_file_multistat(self, fetch_results):
        # Define paths to rejected files
        tmpdir = os.path.dirname(self._f1.name)
        rejdir = os.path.join(tmpdir, 'copylocal1')
        rej1 = os.path.join(rejdir, 'copy_rej.txt')
        except1 = os.path.join(rejdir, 'copy_exception.txt')
        if os.path.isdir(rejdir):
            shutil.rmtree(rejdir)
        rej2 = os.path.join(tmpdir, 'copy_rej2.txt')
        except2 = os.path.join(tmpdir, 'copy_exception2.txt')
        for f in (rej2, except2):
            if os.path.isfile(f):
                os.remove(f)

        # Execute the COPY LOCAL statement as the first and later statement
        # within a query containing multiple statements
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {0} (a INT, b VARCHAR(9))".format(self._table))
            cur.execute(
                "COPY {} FROM LOCAL '{}','{}' DELIMITER ',' ENFORCELENGTH"
                " REJECTED DATA '{}' EXCEPTIONS '{}';"
                "SELECT 100;"
                "COPY {} FROM LOCAL '{}','{}' DELIMITER ',' ENFORCELENGTH"
                " REJECTED DATA '{}' EXCEPTIONS '{}';"
                "SELECT 200;"
                .format(self._table, self._f1.name, self._f2.name, rej1, except1,
                        self._table, self._f3.name, self._f4.name, rej2, except2)
            )

            if fetch_results:
                res1 = cur.fetchall()
                self.assertListOfListsEqual(res1, [[4]])
                self.assertTrue(cur.nextset())

                res2 = cur.fetchall()
                self.assertListOfListsEqual(res2, [[100]])
                self.assertTrue(cur.nextset())

                res3 = cur.fetchall()
                self.assertListOfListsEqual(res3, [[3]])
                self.assertTrue(cur.nextset())

                res4 = cur.fetchall()
                self.assertListOfListsEqual(res4, [[200]])
                self.assertFalse(cur.nextset())

            # There should be no hang/error in the next execution call
            cur.execute("SELECT * FROM {0} ORDER BY a ASC".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[None, 'baz'], [1, 'foo'], [2, 'bar'],
                    [4, None], [11, 'qux'], [13, 'flob'], [15, 'xyz']])

        # Check rejected files
        with open(rej1, 'r', encoding='utf-8') as f:
            self.assertEqual(f.read(), u'x\u00f1,bla\n5,aaaaaaaaaa\n')
        with open(except1, 'r', encoding='utf-8') as f:
            content = f.read()
            self.assertTrue(u"Invalid integer format 'x\u00f1' for column 1 (a)" in content)
            self.assertTrue(u"The 10-byte value is too long for type Varchar(9), column 2 (b)" in content)
        with open(rej2, 'r', encoding='utf-8') as f:
            self.assertEqual(f.read(), u'10,kkkkkkkkkkkk\nxx,corge\nf,quux\n')
        with open(except2, 'r', encoding='utf-8') as f:
            content = f.read()
            self.assertTrue(u"The 12-byte value is too long for type Varchar(9), column 2 (b)" in content)
            self.assertTrue(u"Invalid integer format 'xx' for column 1 (a)" in content)
            self.assertTrue(u"Invalid integer format 'f' for column 1 (a)" in content)

        # Delete files
        try:
            for f in (rej2, except2):
                os.remove(f)
            shutil.rmtree(rejdir)
        except Exception:
            pass

    def test_copy_local_returnrejected(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {0} (a INT, b VARCHAR(9))".format(self._table))
            msg = "COPY LOCAL does not support rejected row numbers with exceptions or rejected data options"
            with pytest.raises(errors.QueryError, match=msg):
                cur.execute(
                    "COPY {} FROM LOCAL '{}','{}' DELIMITER ',' ENFORCELENGTH"
                    " RETURNREJECTED"
                    " REJECTED DATA 'copy_rej.txt' EXCEPTIONS 'copy_exp.txt'"
                    .format(self._table, self._f1.name, self._f2.name))
            # Rejected row numbers write to log file
            cur.execute(
                "COPY {} FROM LOCAL '{}','{}' DELIMITER ',' ENFORCELENGTH"
                " RETURNREJECTED"
                .format(self._table, self._f1.name, self._f2.name))
            self.assertListOfListsEqual(cur.fetchall(), [[4]])
            cur.execute("SELECT * FROM {0} ORDER BY a ASC".format(self._table))
            self.assertListOfListsEqual(cur.fetchall(), [[None, 'baz'], [1, 'foo'], [2, 'bar'], [4, None]])

    def test_copy_local_rejected_as_table(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {0} (a INT, b VARCHAR(9))".format(self._table))
            cur.execute("DROP TABLE IF EXISTS test_loader_rejects CASCADE")
            cur.execute(
                "COPY {} FROM LOCAL '{}','{}' DELIMITER ',' ENFORCELENGTH"
                " REJECTED DATA AS TABLE test_loader_rejects"
                .format(self._table, self._f1.name, self._f2.name))
            self.assertListOfListsEqual(cur.fetchall(), [[4]])

            cur.execute("SELECT rejected_data, rejected_reason FROM test_loader_rejects ORDER BY row_number ASC")
            self.assertListOfListsEqual(cur.fetchall(),
                [['5,aaaaaaaaaa', 'The 10-byte value is too long for type Varchar(9), column 2 (b)'],
                 [u'x\u00f1,bla', u"Invalid integer format 'x\u00f1' for column 1 (a)"]])

            cur.execute("SELECT * FROM {0} ORDER BY a ASC".format(self._table))
            self.assertListOfListsEqual(cur.fetchall(), [[None, 'baz'], [1, 'foo'], [2, 'bar'], [4, None]])

            cur.execute("DROP TABLE IF EXISTS test_loader_rejects CASCADE")

class SimpleQueryExecutemanyTestCase(VerticaPythonIntegrationTestCase):
    def setUp(self):
        super(SimpleQueryExecutemanyTestCase, self).setUp()
        self._table = 'executemany_test'
        self._init_table()

    def _init_table(self):
        with self._connect() as conn:
            cur = conn.cursor()
            # clean old table
            cur.execute("DROP TABLE IF EXISTS {0}".format(self._table))

            # create test table
            cur.execute("""CREATE TABLE {0} (
                                a INT,
                                b VARCHAR(32)
                           )
                        """.format(self._table))

    def tearDown(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("DROP TABLE IF EXISTS {0}".format(self._table))
        super(SimpleQueryExecutemanyTestCase, self).tearDown()

    def _test_executemany(self, table, seq_of_values):
        with self._connect() as conn:
            cur = conn.cursor()

            cur.executemany("INSERT INTO {0} (a, b) VALUES (%s, %s)".format(table),
                            seq_of_values)
            conn.commit()

            cur.execute("SELECT * FROM {0} ORDER BY a ASC, b ASC".format(table))

            # check first select results
            res1 = cur.fetchall()
            seq_of_values_to_compare = sorted([list(values) for values in seq_of_values])
            self.assertListOfListsEqual(res1, seq_of_values_to_compare)
            self.assertIsNone(cur.fetchone())

    def test_executemany(self):
        self._test_executemany(self._table, [(1, 'aa'), (2, 'bb')])

    def test_executemany_quoted_path(self):
        table = '.'.join(['"{}"'.format(s.strip('"')) for s in self._table.split('.')])
        self._test_executemany(table, [(1, 'aa'), (2, 'bb')])

    def test_executemany_utf8(self):
        self._test_executemany(self._table, [(1, u'a\xfc'), (2, u'bb')])

    # test for #292
    def test_executemany_autocommit(self):
        with self._connect() as conn:
            cur = conn.cursor()
            conn.autocommit = False
            cur.execute('BEGIN')
            cur.executemany("INSERT INTO {0} (a, b) VALUES (%s, %s)".format(self._table),
                            ((None, 'foo'), [2, None], [3, 'bar']))
            cur.execute('ROLLBACK')

            cur.execute("SELECT count(*) FROM {0}".format(self._table))
            res = cur.fetchone()[0]
            self.assertEqual(res, 0)

    def test_executemany_null(self):
        seq_of_values_1 = ((None, 'foo'), [2, None])
        seq_of_values_2 = ({'a': None, 'b': 'bar'}, {'a': 4, 'b': None})
        seq_of_values_to_compare = [[None, 'bar'], [None, 'foo'], [2, None], [4, None]]
        with self._connect() as conn:
            cur = conn.cursor()

            cur.executemany("INSERT INTO {0} (a, b) VALUES (%s, %s)".format(self._table),
                            seq_of_values_1)
            conn.commit()
            cur.executemany("INSERT INTO {0} (a, b) VALUES (:a, :b)".format(self._table),
                            seq_of_values_2)
            conn.commit()

            cur.execute("SELECT * FROM {0} ORDER BY a ASC, b ASC".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, seq_of_values_to_compare)
            self.assertIsNone(cur.fetchone())

    def test_executemany_empty(self):
        err_msg = "executemany is implemented for simple INSERT statements only"
        with self._connect() as conn:
            cur = conn.cursor()
            with pytest.raises(NotImplementedError, match=err_msg):
                cur.executemany("", [])


class PreparedStatementTestCase(VerticaPythonIntegrationTestCase):
    @classmethod
    def setUpClass(cls):
        super(PreparedStatementTestCase, cls).setUpClass()
        cls._conn_info['use_prepared_statements'] = True

    def setUp(self):
        super(PreparedStatementTestCase, self).setUp()
        self._table = 'preparedstmt_test'
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("DROP TABLE IF EXISTS {0}".format(self._table))

    def tearDown(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("DROP TABLE IF EXISTS {0}".format(self._table))
        super(PreparedStatementTestCase, self).tearDown()

    def test_empty_statement(self):
        with self._connect() as conn:
            cur = conn.cursor()
            err_msg = 'The statement being prepared is empty'
            with pytest.raises(errors.EmptyQueryError, match=err_msg):
                cur.execute("")
            with pytest.raises(errors.EmptyQueryError, match=err_msg):
                cur.executemany("", [()])
            with pytest.raises(errors.EmptyQueryError, match=err_msg):
                cur.execute("--select 1")
            with pytest.raises(errors.EmptyQueryError, match=err_msg):
                cur.execute("""
                        /*
                        Test block comment
                        */
                        """)

    def test_compound_statement(self):
        with self._connect() as conn:
            cur = conn.cursor()
            query = "select 1; select 2; select 3;"
            err_msg = 'Cannot insert multiple commands into a prepared statement'
            with pytest.raises(errors.VerticaSyntaxError, match=err_msg):
                cur.execute(query)

    def test_num_of_parameters(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {} (a int, b varchar)".format(self._table))
            err_msg = 'Invalid number of parameters'

            # The number of parameters is less than columns of table
            with pytest.raises(ValueError, match=err_msg):
                cur.execute("INSERT INTO {} VALUES (?,?)".format(self._table))

            # The number of parameters is greater than columns of table
            with pytest.raises(ValueError, match=err_msg):
                cur.execute("INSERT INTO {} VALUES (?,?)".format(self._table), [1, 'foo', 2])

            # The number of parameters is equal to columns of table
            values = [1, 'varchar']
            cur.execute("INSERT INTO {} VALUES (?,?)".format(self._table), values)
            conn.commit()
            cur.execute("SELECT * FROM {}".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [values])

    def test_format_paramstyle(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {} (a int, b varchar)".format(self._table))
            err_msg = 'Syntax error at or near "%"'
            values = [1, 'varchar']
            with pytest.raises(errors.VerticaSyntaxError, match=err_msg):
                cur.execute("INSERT INTO {} VALUES (%s, %s)".format(self._table), values)

            cur.execute("INSERT INTO {} VALUES (%s, %s)".format(self._table),
                        values, use_prepared_statements=False)
            conn.commit()
            cur.execute("SELECT * FROM {}".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [values])

    def test_named_paramstyle(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {} (a int, b varchar)".format(self._table))
            err_msg = 'Execute parameters should be a list/tuple'
            values = {'a': 1, 'b': 'varchar'}
            with pytest.raises(TypeError, match=err_msg):
                cur.execute("INSERT INTO {} VALUES (:a, :b)".format(self._table), values)

            cur.execute("INSERT INTO {} VALUES (:a, :b)".format(self._table),
                        values, use_prepared_statements=False)
            conn.commit()
            cur.execute("SELECT * FROM {}".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [[1, 'varchar']])

    def test_executemany(self):
        values = ((None, 'foo'), [1, 'aa'], (2, None), [2, u'a\xfc'])
        expected = [[None, 'foo'], [1, 'aa'], [2, None], [2, u'a\xfc']]
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {} (a int, b varchar)".format(self._table))
            cur.executemany("INSERT INTO {} VALUES (?, ?)".format(self._table), values)

            self.assertListOfListsEqual(cur.fetchall(), [[1]])
            self.assertIsNone(cur.fetchone())
            self.assertTrue(cur.nextset())

            self.assertListOfListsEqual(cur.fetchall(), [[1]])
            self.assertIsNone(cur.fetchone())
            self.assertTrue(cur.nextset())

            self.assertListOfListsEqual(cur.fetchall(), [[1]])
            self.assertIsNone(cur.fetchone())
            self.assertTrue(cur.nextset())

            self.assertListOfListsEqual(cur.fetchall(), [[1]])
            self.assertIsNone(cur.fetchone())
            self.assertFalse(cur.nextset())

            cur.executemany("SELECT * FROM {} WHERE a=? ORDER BY b"
                            .format(self._table), [[1], [2], [3]])
            self.assertListOfListsEqual(cur.fetchall(), [[1, 'aa']])
            self.assertIsNone(cur.fetchone())
            self.assertTrue(cur.nextset())

            self.assertListOfListsEqual(cur.fetchall(), [[2, u'a\xfc'], [2, None]])
            self.assertIsNone(cur.fetchone())
            self.assertTrue(cur.nextset())

            self.assertListOfListsEqual(cur.fetchall(), [])
            self.assertIsNone(cur.fetchone())
            self.assertFalse(cur.nextset())

    def test_bind_boolean(self):
        values = (True, 't', 'true', '1', 1, 'Yes', 'y', None,
                  False, 'f', 'false', '0', 0, 'No', 'n')
        expected = [[True] * 7 + [None] + [False] * 7]
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("""CREATE TABLE {} (
                    col_1 BOOL, col_2 BOOL, col_3 BOOL, col_4 BOOL, col_5 BOOL,
                    col_6 BOOL, col_7 BOOL, col_8 BOOL, col_9 BOOL, col_10 BOOL,
                    col_11 BOOL, col_12 BOOL, col_13 BOOL, col_14 BOOL, col_15 BOOL
                    )""".format(self._table))
            cur.execute("INSERT INTO {} VALUES ({})".format(
                        self._table, ','.join(['?']*15)), values)
            conn.commit()
            cur.execute("SELECT * FROM {}".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, expected)

    def test_bind_datetime(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {} (a TIMESTAMP, b DATE, c TIME, "
                        "d TIMESTAMP, e DATE, f TIME)".format(self._table))
            values = [datetime(2018, 9, 7, 15, 38, 19, 769000), date(2018, 9, 7),
                      time(13, 50, 9), None, None, None]
            cur.execute("INSERT INTO {} VALUES (?,?,?,?,?,?)".format(self._table), values)
            conn.commit()
            cur.execute("SELECT * FROM {}".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, [values])

    def test_bind_binary(self):
        values = [b'binary data', b'\\backslash data\\', u'\\backslash data\\',
                  u'\u00f1 encoding', 'raw data', 'long varbinary data', None]
        expected = [[b'binary data\\000\\000\\000', b'\\\\backslash data\\\\',
                     b'\\\\backslash data\\\\', b'\\303\\261 encoding',
                     b'raw data', b'long varbinary data', None]]
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("""CREATE TABLE {} (
                    col_1 BINARY(14), col_2 VARBINARY, col_3 VARBINARY,
                    col_4 BYTEA, col_5 RAW, col_6 LONG VARBINARY,
                    col_7 VARBINARY)""".format(self._table))
            cur.execute("INSERT INTO {} VALUES ({})".format(
                        self._table, ','.join(['?']*7)), values)
            conn.commit()
            cur.execute("SELECT * FROM {}".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, expected)

    def test_bind_numeric(self):
        values = [Decimal('123456789.98765'), Decimal('123456789.98765'),
                  Decimal('123456789.98765'), Decimal('123456789.98765'),
                  10, 11, 1234567890, 1296554905964, 123, 123, 123.45,
                  123.45, 123.45, 123.45, 123.45, None, None, None]
        expected = [[Decimal('123456789.9877'), Decimal('123456789.987650000000000'),
                     Decimal('123456790'), Decimal('123456789.987650000000000'),
                     10, 11, 1234567890, 1296554905964, 123, 123, 123.45,
                     123.45, 123.45, 123.45, 123.45, None, None, None]]
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("""CREATE TABLE {} (
                col_1 MONEY, col_2 NUMERIC, col_3 NUMBER, col_4 DECIMAL,
                col_5 TINYINT, col_6 SMALLINT, col_7 INT8, col_8 BIGINT,
                col_9 INT, col_10 INTEGER, col_11 DOUBLE PRECISION, col_12 REAL,
                col_13 FLOAT8, col_14 FLOAT(10), col_15 FLOAT,
                col_16 NUMERIC, col_17 INT, col_18 FLOAT
                )""".format(self._table))
            cur.execute("INSERT INTO {} VALUES ({})".format(
                        self._table, ','.join(['?']*18)), values)
            conn.commit()
            cur.execute("SELECT * FROM {}".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, expected)

    def test_bind_character(self):
        values = ['char data', b'raw varchar data',
                  u'long varbinary data \u00f1', None, None, None]
        expected = [['char data ', 'raw varchar data',
                     u'long varbinary data \u00f1', None, None, None]]
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("""CREATE TABLE {} (
                col_1 CHAR(10), col_2 VARCHAR, col_3 LONG VARCHAR,
                col_4 CHAR(10), col_5 VARCHAR, col_6 LONG VARCHAR
                )""".format(self._table))
            cur.execute("INSERT INTO {} VALUES ({})".format(
                        self._table, ','.join(['?']*6)), values)
            conn.commit()
            cur.execute("SELECT * FROM {}".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, expected)

    def test_bind_interval(self):
        values = ['1y 10m', '1 2', '1y 10m', '1y 10m',
                  '17910y 1h 3m 6s 5msecs 57us ago', '3 days 2 hours', '1 3',
                  '1y 15 mins 20 sec', '15 mins 20 sec', '1y 5 mins 20 sec',
                  '2 days 12 hours 15 mins 1235 milliseconds',
                  '2 days 12 hours 15 mins ago', '2 days 12 hours 15 mins ago',
                  None, None]
        expected = [[b'1', b'1-2', b'22', b'365', b'-6537150 01:03:06.0051',
                     b'74', b'01:03', b'8760:15:20', b'15', b'525605:20',
                     b'216901.24', b'-2 12', b'-2 12:15', None, None]]
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("""CREATE TABLE {} (
                col_1 INTERVAL YEAR, col_2 INTERVAL YEAR TO MONTH,
                col_3 INTERVAL MONTH, col_4 INTERVAL DAY,
                col_5 INTERVAL DAY TO SECOND(4), col_6 INTERVAL HOUR,
                col_7 INTERVAL HOUR to MINUTE, col_8 INTERVAL HOUR to SECOND,
                col_9 INTERVAL MINUTE, col_10 INTERVAL MINUTE TO SECOND,
                col_11 INTERVAL SECOND(2), col_12 INTERVAL DAY TO HOUR,
                col_13 INTERVAL DAY TO MINUTE, col_14 INTERVAL YEAR TO MONTH,
                col_15 INTERVAL DAY TO SECOND
                )""".format(self._table))
            cur.execute("INSERT INTO {} VALUES ({})".format(
                        self._table, ','.join(['?']*15)), values)
            conn.commit()
            cur.execute("SELECT * FROM {}".format(self._table))
            res = cur.fetchall()
            self.assertListOfListsEqual(res, expected)

    def test_bind_udtype(self):
        poly = "POLYGON ((1 2, 2 3, 3 1, 1 2))"
        line = "LINESTRING (42.1 71, 41.4 70, 41.3 72.9, 42.99 71.46, 44.47 73.21)"
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute("CREATE TABLE {} (c1 GEOMETRY(10000), c2 GEOGRAPHY(1000))"
                .format(self._table))

            cur.execute("INSERT INTO {} VALUES (ST_GeomFromText(?), ST_GeographyFromText(?))"
                .format(self._table), [poly, line])
            conn.commit()
            cur.execute("SELECT c1, c2, ST_AsText(c1), ST_AsText(c2) FROM {}".format(self._table))

            res = cur.fetchall()
            self.assertEqual(res[0][2], poly)
            self.assertEqual(res[0][3], line)

            datatype_names = [col.type_name for col in cur.description]
            expected = ['geometry', 'geography', 'Long Varchar', 'Long Varchar']
            self.assertListEqual(datatype_names, expected)

            self.assertEqual(cur.description[0].display_size, 10000)
            self.assertEqual(cur.description[1].display_size, 1000)



# Copyright (c) 2020 Micro Focus or one of its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import print_function, division, absolute_import

import os
from struct import pack

from ..message import BulkFrontendMessage


class VerifiedFiles(BulkFrontendMessage):
    message_id = b'F'

    def __init__(self, file_list):
        BulkFrontendMessage.__init__(self)
        self.filenames = file_list

    def read_bytes(self):
        bytes_ = pack('!H', len(self.filenames))
        for filename in self.filenames:
            utf_filename = filename.encode('utf-8')
            bytes_ += pack('!{0}sx'.format(len(utf_filename)), utf_filename)
            bytes_ += pack('!Q', os.path.getsize(filename))

        return bytes_

    def __str__(self):
        return "VerifiedFiles: {}".format(self.filenames)
